# mcsamplerNFlows
#    Based on A.H. Fernando code
#
# Refs
#   https://github.com/bayesiains/nflows/blob/master/examples/conditional_moons.ipynb
#   https://dfdazac.github.io/02-flows.html


import sys
import math
#import bisect
from collections import defaultdict

import numpy
np=numpy #import numpy as np
from scipy import integrate, interpolate, special
import itertools
import functools

import os


from RIFT.integrators.mcsampler_generic import MCSamplerGeneric

from typing import List, Tuple

import torch
from torch import optim

from nflows.flows.base import Flow
from nflows.distributions.normal import StandardNormal
from nflows.transforms.base import CompositeTransform
from nflows.transforms.autoregressive import MaskedAffineAutoregressiveTransform
from nflows.transforms.autoregressive import MaskedPiecewiseLinearAutoregressiveTransform
from nflows.transforms.autoregressive import MaskedPiecewiseQuadraticAutoregressiveTransform
from nflows.transforms.autoregressive import MaskedPiecewiseCubicAutoregressiveTransform
from nflows.transforms.permutations import ReversePermutation
# -------------------------------------
from nflows.transforms.base import InverseTransform
from nflows.transforms.base import MultiscaleCompositeTransform
from nflows.transforms.standard import IdentityTransform
#from nflows.transforms.standard import AffineScalarTransform
#from nflows.transforms.standard import AffineTransform
from nflows.transforms.standard import PointwiseAffineTransform
from nflows.transforms.linear import NaiveLinear

try:
  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
except:
  True


try:
  import cupy
  import cupyx   # needed for logsumexp
  xpy_default=cupy
  try:
    xpy_special_default = cupyx.scipy.special
    if not(hasattr(xpy_special_default,'logsumexp')):
          print(" mcsamplerAV: no cupyx.scipy.special.logsumexp, fallback mode ...")
          xpy_special_default= special
  except:
    print(" mcsamplerAV: no cupyx.scipy.special, fallback mode ...")
    xpy_special_default= special
  identity_convert = cupy.asnumpy
  identity_convert_togpu = cupy.asarray
  junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
  cupy_ok = True
  cupy_pi = cupy.array(np.pi)

  from RIFT.interpolators.interp_gpu import interp

#  from logging import info as log
#  import inspect
#  def verbose_cupy_asarray(*args, **kwargs):
#     print("Transferring data to VRAM", *args, **kwargs)
#     return cupy.asarray(*args, **kwargs)
#  def verbose_cupy_asnumpy(*args, **kwargs):
#     curframe = inspect.currentframe()
#     calframe = inspect.getouterframes(curframe, 2)
#     log("Transferring data to RAM",calframe[1][3]) #,args[0].__name__) #, *args, **kwargs)
#     return cupy.ndarray.asnumpy(*args, **kwargs)
#  cupy.asarray = verbose_cupy_asarray  
#  cupy.ndarray.asnumpy = verbose_cupy_asnumpy

except:
  print(' no cupy (mcsamplerAV)')
#  import numpy as cupy  # will automatically replace cupy calls with numpy!
  xpy_default=numpy  # just in case, to make replacement clear and to enable override
  xpy_special_default = special
  identity_convert = lambda x: x  # trivial return itself
  identity_convert_togpu = lambda x: x
  cupy_ok = False
  cupy_pi = np.pi

def set_xpy_to_numpy():
   xpy_default=numpy
   identity_convert = lambda x: x  # trivial return itself
   identity_convert_togpu = lambda x: x
   cupy_ok = False
   

if 'PROFILE' not in os.environ:
   def profile(fn):
        return fn

if not( 'RIFT_LOWLATENCY'  in os.environ):
    # Dont support selected external packages in low latency
 try:
    import healpy
 except:
    print(" - No healpy - ")

from RIFT.integrators.statutils import  update,finalize, init_log,update_log,finalize_log

#from multiprocessing import Pool

from RIFT.likelihood import vectorized_general_tools

__author__ = "R. O'Shaughnessy, A. H. Fernando"

rosDebugMessages = True

class NanOrInf(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


def check_nonzero_deriv(my_seq,deriv_crit =0.05):
  """
  Computes  dy/dx/y \simeq (<yx> - <x><y>)/(<x><y>)
  Tries to require more than a certain percentage change per test block
  """
  mean_y = np.mean(my_seq)  # <y>
  mean_x  = len(my_seq)/2
  mean_yx = np.mean(my_seq*np.arange(len(my_seq))) # <xy>
  dydx_scale = mean_yx/(mean_y*mean_x) - 1  # relative change
  std_y_scale = np.std(my_seq)/mean_y    # relative noise
#  print(" deriv check ", dydx_scale, np.mean(my_seq), std_y_scale)  # should be PER ITERATION check
  if dydx_scale < -deriv_crit:
    return True


class NFlowsNFS_Trainer:
    def __init__(self, bounds: List[Tuple[float, float]], 
                 n_samples=100,
                 target_distribution= StandardNormal,
                 base_distribution=StandardNormal,
                 transform=CompositeTransform):
        self.bounds              = bounds
        self.n_samples           = n_samples
        self.target_distribution = target_distribution
        self.base_distribution   = base_distribution
        self.transform           = transform
        self.flow                = None
        self.plotting            = False
        self.loss_history = []
    
    def set_plotting(self, plotting: bool = False) -> None:
        self.plotting            = plotting
    
    def train_flow(self, samples_in: List[List[float]],
                   out_n_samples: int, max_epochs: int, bound_offset: float,n_print=50):
        if self.flow is None:
          self.flow                = Flow(self.transform, self.base_distribution(shape=[len(self.bounds)]))

        optimizer                = optim.Adam(self.flow.parameters())
        prev_efficiency          = float('inf')
        losses                   = self.loss_history
        # Fixed training input
        samples              = torch.tensor(samples_in, dtype=torch.float32)
        for epoch in range(1, int(max_epochs)):       
            # Set gradients equals to zero
            optimizer.zero_grad()
            
            # Compute the loss
            loss                 = -self.flow.log_prob(samples).mean()
            
            # Backpropergate the loss
            loss.backward()
            
            # Store loss value for monitoring
            losses.append(loss.item()) 
            if (epoch%n_print)==0:
              print(epoch, loss.item())
              n_hist = np.min([len(losses), 100])
#              if check_nonzero_deriv(losses[-n_hist:]):
#                print("    -- stationary ---")
            
            optimizer.step()
            
            # Display Results Every 500 Iterations 
            if (epoch + 1) % 500 == 0:
                # with torch.no_grad():
                #     min_bounds   = [bound[0] - bound_offset for bound in self.bounds]
                #     max_bounds   = [bound[1] + bound_offset for bound in self.bounds]
                #     xgrid        = [torch.linspace(min_bound, 
                #                                    max_bound, 
                #                                    out_n_samples) for min_bound, 
                #                                                       max_bound in zip(min_bounds, 
                #                                                                        max_bounds)]
                #     meshgrids    = torch.meshgrid(*xgrid)
                #     xyinput      = torch.cat([meshgrid.reshape(-1, 1) for meshgrid in meshgrids], dim=1)

                #     zgrid        = self.flow.log_prob(xyinput).exp().reshape([out_n_samples] * len(self.bounds))
                    
                # if len(self.bounds) == 2 and self.plotting == True:
                #     self.plot_if_2d(epoch, meshgrids, zgrid, loss)
              #else:
              print(f"Iteration: {epoch + 1}   Training Loss: {loss}")

        return losses
    
    
    def plot_if_2d(self, epoch, meshgrids, zgrid, loss) -> None:
        plt.figure()
        dummy_variable = plt.contourf(*meshgrids, zgrid.numpy())
        plt.title(f"Iteration: {epoch + 1}   Training Loss: {loss}")
        plt.savefig("fig_{}.png".format(epoch))




class MCSampler(MCSamplerGeneric):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """


    def __init__(self,n_chunk=400000,**kwargs):
        # Total number of samples drawn
        self.ntotal = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        # If the pdfs aren't normalized, this will hold the normalization 
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        # params for left and right limits
        self.llim, self.rlim = {}, {}


        self.n_chunk = n_chunk
        self.nbins = None
        self.ninbin = None
        self.adaptive =[]

        self.pdf = {} # not used

        # MEASURES (=priors): ROS needs these at the sampler level, to clearly separate their effects
        # ASSUMES the user insures they are normalized
        self.prior_pdf = {}

        # histogram setup
        self.xpy = numpy
        self.identity_convert = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)

        # sampling tool
        self.nf_model = None
        self.nf_trainer = None
        self.nf_flow = None


    def setup(self, nf_cov=None, nf_mean=None,nf_method=None,**kwargs):

        self._rvs={}
        self.lnL_thresh = -np.inf
        self.enc_prob = 0.999
        d_nf = len(self.params_ordered)

        bounds = np.array([ [self.llim[p],self.rlim[p]] for p in self.params_ordered])

        # https://github.com/bayesiains/nflows/blob/master/examples/moons.ipynb
        self.num_layers = int(len(bounds)/2)  # autoregressive
        transforms = []
        # trivial scale layer first, to get in the right place
#        transforms.append(PointwiseAffineTransform())
#        transforms.append(NaiveLinear(features=len(bounds)))
        for _ in range(self.num_layers):
          transforms.append(ReversePermutation(features = len(bounds)))
          transforms.append(MaskedAffineAutoregressiveTransform(features = len(bounds),
                                                          hidden_features = 2 * len(bounds)))    
        transform  = CompositeTransform(transforms)
        self.nf_model = transform

        trainer    = NFlowsNFS_Trainer(bounds              = bounds, 
                               transform           = transform)
        self.nf_trainer = trainer
        self.nf_flow = trainer.flow
        


    def clear(self):
        """
        Clear out the parameters and their settings, as well as clear the sample cache.
        """
        self.params = set()
        self.params_ordered = []
        self.pdf = {}
        self._pdf_norm = defaultdict(lambda: 1.0)
        self._rvs = {}
        self.llim = {}
        self.rlim = {}
        self.adaptive = []


    def add_parameter(self, params, pdf,  cdf_inv=None, left_limit=None, right_limit=None, prior_pdf=None, adaptive_sampling=False):
        """
        Add one (or more) parameters to sample dimensions. params is either a string describing the parameter, or a tuple of strings. The tuple will indicate to the sampler that these parameters must be sampled together. left_limit and right_limit are on the infinite interval by default, but can and probably should be specified. If several params are given, left_limit, and right_limit must be a set of tuples with corresponding length. Sampling PDF is required, and if not provided, the cdf inverse function will be determined numerically from the sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        if rosDebugMessages: 
            print(" Adding parameter ", params, " with limits ", [left_limit, right_limit])
        if isinstance(params, tuple):
            assert all([lim[0] < lim[1] for lim in zip(left_limit, right_limit)])
            if left_limit is None:
                self.llim[params] = list(float("-inf"))*len(params)
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = list(float("+inf"))*len(params)
            else:
                self.rlim[params] = right_limit
        else:
            assert left_limit < right_limit
            if left_limit is None:
                self.llim[params] = float("-inf")
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = float("+inf")
            else:
                self.rlim[params] = right_limit
        self.pdf[params] = pdf
        self.prior_pdf[params] = prior_pdf

        if adaptive_sampling:
            print("   Adapting ", params)
            self.adaptive.append(params)

    def prior_prod(self, x):
        """
        Evaluates prior_pdf(x), multiplying together all factors
        """
        p_out = np.ones(len(x))
        indx = 0
        for param in self.params_ordered:
            p_out *= self.prior_pdf[param](x[:,indx])
            indx +=1
        return p_out


    def draw_simplified(self,n_to_get, *args, **kwargs):
        verbose = kwargs["verbose"] if "verbose" in kwargs else False  # default
        super_verbose = kwargs["super_verbose"] if "super_verbose" in kwargs else False  # default
        save_no_samples = kwargs.get("save_no_samples", False)
        enforce_bounds = kwargs["enforce_bounds"] if "enforce_bounds" in kwargs else False

        args = self.params_ordered # by default draw all

        if self.nf_flow is None:
          if super_verbose:
            print(" Using uniform ")
          bounds = np.array([ [self.llim[p],self.rlim[p]] for p in self.params_ordered])
          V = np.prod((bounds[:,1] - bounds[:,0]))
          # rv = np.random.uniform(low=bounds[:,0], high=bounds[:,1], size = n_to_get)
          rv = np.empty(( len(self.params_ordered),n_to_get))
          for indx, p in enumerate(self.params_ordered):
            rv[indx] = np.random.uniform(self.llim[p], self.rlim[p], size=n_to_get)
          log_ps = -np.log(V)*np.ones(n_to_get)  # de facto constaint on hypercube
          log_p = np.log(self.prior_prod(rv.T))
#          print('ps shape', log_p.shape)
        else:
          if super_verbose:
            print(" Using actual flow ")
          flow = self.nf_flow
          flow_samples = flow.sample(n_to_get)
          rv = flow_samples.detach().numpy().T
          log_ps = flow.log_prob(flow_samples).detach().numpy()
          log_p  =  np.log(self.prior_prod(rv.T))
          # remove nan values
          # enforce boundaries: don't trust flow
          if enforce_bounds:
            indx_valid = np.ones(len(log_p), dtype=bool)
            bounds = np.array([ [self.llim[p],self.rlim[p]] for p in self.params_ordered])
            for indx, p in enumerate(self.params_ordered):
              indx_valid = np.logical_and(indx_valid, rv[indx] <= self.rlim[p])
              indx_valid = np.logical_and(indx_valid, rv[indx] >= self.llim[p])
            if super_verbose:
              print(" Valid ", np.sum(indx_valid))
            rv = rv[:,indx_valid]
            log_ps = log_ps[indx_valid]
            log_p = log_p[indx_valid]

        # Cache the samples we chose
        #
        if not save_no_samples:
#            print(" ===== RECORDING SAMPLES ====")
            if len(self._rvs) == 0:
               self._rvs = dict(list(zip(args, rv)))
#               print(self._rvs)
            else:
               rvs_tmp = dict(list(zip(args, rv)))
#               print(rvs_tmp)
               #for p, ar in self._rvs.items():
               for p in self.params_ordered:
                   self._rvs[p] = self.xpy.hstack((self._rvs[p], rvs_tmp[p]))


        return  rv, np.exp(log_ps), np.exp(log_p)

    def update_sampling_prior(self, lnw, *args, xpy=xpy_default,no_protect_names=True,external_rvs=None,tempering_exp=1,n_history=1000,**kwargs):
      """
      update_sampling_prior

      """
      verbose = kwargs["verbose"] if "verbose" in kwargs else False  # default
      super_verbose = kwargs["super_verbose"] if "super_verbose" in kwargs else False  # default

      xpy_here = self.xpy
      enforce_bounds=True

      # Allow updates provided from outside sources,
      rvs_here = self._rvs
      if external_rvs:
        rvs_here = external_rvs

      # skip update if lnw is too flat
      if np.mean(lnw) > np.max(lnw)-1.5*len(self.params_ordered):
        if super_verbose:
          print(" Skipping update ")
        return 

      # apply tempering exponent (structurally slightly different than in low-level code - not just to likelihood)
      ln_weights  = self.xpy.array(lnw) # force copy
      ln_weights *= tempering_exp

      n_history_to_use = np.min([n_history, len(ln_weights), len(rvs_here[self.params_ordered[0]])] )
      if (n_history_to_use) < 10:
        print("  Skipping update: no history ")
        return

      # default is to use logarithmic (!) weights, relying on them being positive.
      weights_alt = ln_weights[-n_history_to_use:]  - self.xpy.max(ln_weights) + 100  
#      weights_alt = self.xpy.maximum(weights_alt, 1e-5)    # prevent negative weights, in case integrating function with lnL < 0
      # now treat as sum
      weights_alt = weights_alt/(weights_alt.sum())
      if weights_alt.dtype == numpy.float128:
        weights_alt = weights_alt.astype(numpy.float64,copy=False)


      # copy into training array, truncating size
      samples_train = np.zeros((len(self.params_ordered),n_history_to_use))
      for itr, p in enumerate(self.params_ordered):
                samples_train[itr] = rvs_here[p][-n_history_to_use:]

      # Eliminate points with low weight.  Drop bottom half, right now
      indx_ok = weights_alt >  np.mean(weights_alt) +0.5*np.std(weights_alt)     
      if np.sum(indx_ok) < 10:
        if super_verbose:
          print(" Skipping update: too few valid ") 
        return
#      print(indx_ok.shape, samples_train.shape)
      samples_train = samples_train[:,indx_ok]
      if samples_train.shape[1] < 10:
        if super_verbose:
          print(" Skipping update: too few valid ") 
        return        
      if super_verbose:
        print("    NF: Training data shape ", samples_train.shape)
        print("    NF: training data mean ", np.mean(samples_train,axis=-1))

      # Train
      trainer = self.nf_trainer
      max_epochs = 300  # be short, don't need precision/long training
      if max_epochs < int(10*self.num_layers): max_epochs = int(10*self.num_layers)

      losses     = trainer.train_flow(samples_in= samples_train.T,
                                out_n_samples = 100, 
                                max_epochs    = max_epochs, 
                                bound_offset  = 0.5)

      self.nf_flow = trainer.flow


    @profile
    def integrate_log(self, lnF, *args, xpy=xpy_default,**kwargs):
        """
        Integrate exp(lnF) returning lnI, by using n sample points, assuming integrand is lnF
        Does NOT allow for tuples of arguments, an unused feature in mcsampler

        tempering is done with lnF, suitably modified.

        kwargs:
        nmax -- total allowed number of sample points, will throw a warning if this number is reached before neff.
        neff -- Effective samples to collect before terminating. If not given, assume infinity
        n -- Number of samples to integrate in a 'chunk' -- default is 1000
        save_integrand -- Save the evaluated value of the integrand at the sample points with the sample point
        history_mult -- Number of chunks (of size n) to use in the adaptive histogramming: only useful if there are parameters with adaptation enabled
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings
        temper_log -- Adapt in min(ln L, 10^(-5))^tempering_exp
        tempering_adapt -- Gradually evolve the tempering_exp based on previous history.
        floor_level -- *total probability* of a uniform distribution, averaged with the weighted sampled distribution, to generate a new sampled distribution
        n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
        convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING
        Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
        """


        xpy_here = self.xpy
        
        #
        # Determine stopping conditions
        #
        nmax = kwargs["nmax"] if "nmax" in kwargs else float("inf")
        neff = kwargs["neff"] if "neff" in kwargs else numpy.float128("inf")
        n = int(kwargs["n"] if "n" in kwargs else min(100000, nmax))
        convergence_tests = kwargs["convergence_tests"] if "convergence_tests" in kwargs else None
        save_no_samples = kwargs["save_no_samples"] if "save_no_samples" in kwargs else None


        #
        # Adaptive sampling parameters
        #
        n_history = int(kwargs["history_mult"]*n) if "history_mult" in kwargs else 2*n
        if n_history<=0:
            print("  Note: cannot adapt, no history ")

        tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0
        n_adapt = int(kwargs["n_adapt"]) if "n_adapt" in kwargs else 10  # default to adapt to 10 chunks, then freeze
        floor_integrated_probability = kwargs["floor_level"] if "floor_level" in kwargs else 0
        temper_log = kwargs["tempering_log"] if "tempering_log" in kwargs else False
        tempering_adapt = kwargs["tempering_adapt"] if "tempering_adapt" in kwargs else False
            

        save_intg = kwargs["save_intg"] if "save_intg" in kwargs else False
        # FIXME: The adaptive step relies on the _rvs cache, so this has to be
        # on in order to work
        if n_adapt > 0 and tempering_exp > 0.0:
            save_intg = True

        deltalnL = kwargs['igrand_threshold_deltalnL'] if 'igrand_threshold_deltalnL' in kwargs else float("Inf") # default is to return all
        deltaP    = kwargs["igrand_threshold_p"] if 'igrand_threshold_p' in kwargs else 0 # default is to omit 1e-7 of probability
        bFairdraw  = kwargs["igrand_fairdraw_samples"] if "igrand_fairdraw_samples" in kwargs else False
        n_extr = kwargs["igrand_fairdraw_samples_max"] if "igrand_fairdraw_samples_max" in kwargs else None

        bShowEvaluationLog = kwargs['verbose'] if 'verbose' in kwargs else True
        bShowEveryEvaluation = kwargs['extremely_verbose'] if 'extremely_verbose' in kwargs else False


        verbose = kwargs["verbose"] if "verbose" in kwargs else False  # default
        super_verbose = kwargs["super_verbose"] if "super_verbose" in kwargs else False  # default
        dict_return_q = kwargs["dict_return"] if "dict_return" in kwargs else False  # default.  Method for passing back rich data structures for debugging

        # use integer numbers of bins always
        enforce_bounds = kwargs["enforce_bounds"] if "enforce_bounds" in kwargs else False

        if bShowEvaluationLog:
            print(" .... mcsampler : providing verbose output ..... ")

        current_log_aggregate = None
        eff_samp = 0  # ratio of max weight to sum of weights
        maxlnL = -np.inf  # max lnL
        maxval=0   # max weight
        outvals=None  # define in top level scope
        self.ntotal = 0
        if bShowEvaluationLog:
            print("iteration Neff  sqrt(2*lnLmax) sqrt(2*lnLmarg) ln(Z/Lmax) int_var")

        self.n_chunk = n
        self.setup()  # sets up self.my_ranges, self.dx initially

        ntotal_true = 0
        while (eff_samp < neff and ntotal_true < nmax ): #  and (not bConvergenceTests):
            # Draw samples. Note state variables binunique, ninbin -- so we can re-use the sampler later outside the loop
            rv, joint_p_s, joint_p_prior = self.draw_simplified(self.n_chunk, save_no_samples=False)  # Beware reversed order of rv
            if super_verbose:
              print(" Drawn ", np.mean(rv, axis=-1))
#              print(" Drawn ", np.cov(rv))
#            print(" Drawn rv ", rv)
#            print(self._rvs)
            log_joint_p_s = np.log(joint_p_s)
            log_joint_p_prior = np.log(joint_p_prior)
            ntotal_true += len(joint_p_s)
            if cupy_ok:
              rv = identity_convert_togpu(rv) # send random numbers to GPU : ugh
              log_joint_p_prior = identity_convert_togpu(log_joint_p_prior)    # send to GPU if required. Don't waste memory reassignment otherwise

            # Evaluate function, protecting argument order
            params = []
            for item in self.params_ordered:  # USE IN ORDER
                if isinstance(item, tuple):
                    params.extend(item)
                else:
                    params.append(item)
            unpacked = unpacked0 = rv #numpy.hstack([r.flatten() for r in rv]).reshape(len(args), -1)
            unpacked = dict(list(zip(params, unpacked)))
            if 'no_protect_names' in kwargs:
                lnL = lnF(*unpacked0)  # do not protect order
            else:
                unpacked = dict(list(zip(self.params_ordered,rv.T)))
                lnL= lnF(**unpacked)  # protect order using dictionary
            # take log if we are NOT using lnL
            if cupy_ok:
              if not(isinstance(lnL,cupy.ndarray)):
                lnL = identity_convert_togpu(lnL)  # send to GPU, if not already there


            # For now: no prior, just duplicate VT algorithm
#            print(rv.shape, lnL.shape, log_joint_p_prior.shape,log_joint_p_s.shape)
            log_integrand =lnL  + log_joint_p_prior - log_joint_p_s
#            log_weights = tempering_exp*lnL + log_joint_p_prior
            # log aggregate: NOT USED at present, remember the threshold is floating
            if current_log_aggregate is None:
              current_log_aggregate = init_log(log_integrand,xpy=xpy,special=xpy_special_default)
            else:
              current_log_aggregate = update_log(current_log_aggregate, log_integrand,xpy=xpy,special=xpy_special_default)

            # Adapt if needed; decrement adaptation counter
            if n_adapt > 0:
              if super_verbose:
                print("    -- n_adapt {} ".format(n_adapt))
              self.update_sampling_prior(lnL,**kwargs)
              n_adapt += -1  # decrement
            else:
              if super_verbose:
                print("  ... skipping adaptation (NF) ")

            # Monitoring for i/o
            outvals = finalize_log(current_log_aggregate,xpy=xpy)
            self.ntotal = current_log_aggregate[0]
            # effective samples
            maxval = max(maxval, identity_convert(self.xpy.max(log_integrand) ))

            # sum of weights is the integral * the number of points
            eff_samp = xpy.exp(  outvals[0]+np.log(self.ntotal) - maxval)   # integral value minus floating point, which is maximum
            if bShowEvaluationLog:
                print(" :",  self.ntotal, eff_samp, numpy.sqrt(2*maxlnL), numpy.sqrt(2*outvals[0]), outvals[0]-maxlnL, np.exp(outvals[1]/2  - outvals[0]  - np.log(self.ntotal)/2 ))

            if save_intg:
                # FIXME: See warning at beginning of function. The prior values
                # need to be moved out of this, as they are not part of MC
                # integration
                if "log_integrand" in self._rvs:
                    self._rvs["log_integrand"] = xpy_here.hstack( (self._rvs["log_integrand"], lnL) )
                    self._rvs["log_joint_prior"] = xpy_here.hstack( (self._rvs["log_joint_prior"], self.xpy.log(joint_p_prior)) )
                    self._rvs["log_joint_s_prior"] = xpy_here.hstack( (self._rvs["log_joint_s_prior"], self.xpy.log(joint_p_s)))
                else:
                    self._rvs["log_integrand"] = lnL
                    self._rvs["log_joint_prior"] = self.xpy.log(joint_p_prior)
                    self._rvs["log_joint_s_prior"] = self.xpy.log(joint_p_s)



        # Manual estimate of integrand, done transparently (no 'log aggregate' or running calculation -- so memory hog
        log_wt = self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"]
        log_wt = identity_convert(log_wt)  # convert to CPU
        log_int = special.logsumexp( log_wt) - np.log(len(log_wt))  # mean value
        rel_var = np.var( np.exp(log_wt - log_int))/len(log_wt)   # error in integral, estimated: just taking int = <w> , so error is V(w_k)/N (sample mean/variance)
        eff_samp = np.sum(np.exp(log_wt - np.max(log_wt)))

        # Integral value: NOT RELIABLE b/c not just using samples in 
#        outvals = finalize_log(current_log_aggregate,xpy=xpy)
#        log_wt_tmp = allloglkl[np.isfinite(allloglkl)]  # remove infinite entries
#        outvals = init_log(log_wt_tmp)
#        print(outvals, log_int, maxval, current_log_aggregate)
#        eff_samp = xpy.exp(  outvals[0]+np.log(len(allloglkl)) - maxval)   # integral value minus floating point, which is maximum
#        rel_var = np.exp(outvals[1]/2  - outvals[0]  - np.log(self.ntotal)/2 )

        # Do a fair draw of points, if option is set. CAST POINTS BACK TO NUMPY, IDEALLY
        if bFairdraw and not(n_extr is None):
           n_extr = int(numpy.min([n_extr,1.5*identity_convert(eff_samp),1.5*neff]))
           print(" Fairdraw size : ", n_extr)
           ln_wt = self.xpy.array(self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"] ,dtype=float)
           ln_wt = identity_convert(ln_wt)  # send to CPU
           ln_wt += - special.logsumexp(ln_wt)
           wt = xpy.exp(identity_convert_togpu(ln_wt))
           if n_extr < len(self._rvs["log_integrand"]):
               indx_list = self.xpy.random.choice(self.xpy.arange(len(wt)), size=n_extr,replace=True,p=wt) # fair draw
               # FIXME: See previous FIXME
               for key in list(self._rvs.keys()):
                   if isinstance(key, tuple):
                       self._rvs[key] = identity_convert(self._rvs[key][:,indx_list])
                   else:
                       self._rvs[key] = identity_convert(self._rvs[key][indx_list])


        # perform type conversion of all stored variables.  VERY LARGE -- should only do this if we need it!
        if cupy_ok:
          for name in self._rvs:
            if isinstance(self._rvs[name],xpy_default.ndarray):
              self._rvs[name] = identity_convert(self._rvs[name])   # this is trivial if xpy_default is numpy, and a conversion otherwise

        dict_return = {}
        return log_int, np.log(rel_var)  +2*log_int, eff_samp, dict_return

        # if outvals:
        #   out0 = outvals[0]; out1 = outvals[1]
        #   if not(isinstance(outvals[0], np.float64)):
        #     # type convert everything as needed
        #     out0 = identity_convert(out0)
        #   if not(isinstance(outvals[1], np.float64)):
        #     out1 = identity_convert(out1)
        #     eff_samp = identity_convert(eff_samp)
        #   return out0, out1 - np.log(self.ntotal), eff_samp, dict_return
        # else: # very strange case where we terminate early
        #   return None, None, None, None


    @profile
    def integrate(self, func, *args, **kwargs):
        """
        Integrate func, by using n sample points. Right now, all params defined must be passed to args must be provided, but this will change soon.
        Does NOT allow for tuples of arguments, an unused feature in mcsampler

        kwargs:
        nmax -- total allowed number of sample points, will throw a warning if this number is reached before neff.
        neff -- Effective samples to collect before terminating. If not given, assume infinity
        n -- Number of samples to integrate in a 'chunk' -- default is 1000
        save_integrand -- Save the evaluated value of the integrand at the sample points with the sample point
        history_mult -- Number of chunks (of size n) to use in the adaptive histogramming: only useful if there are parameters with adaptation enabled
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings
        temper_log -- Adapt in min(ln L, 10^(-5))^tempering_exp
        tempering_adapt -- Gradually evolve the tempering_exp based on previous history.
        floor_level -- *total probability* of a uniform distribution, averaged with the weighted sampled distribution, to generate a new sampled distribution
        n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
        convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING
        Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
        """
        def ln_func(*args):
          return np.log(func(*args))
        infunc = ln_func
        use_lnL=False
        if 'use_lnL' in kwargs:   # should always be positive
          if kwargs['use_lnL']:
            infunc = func
            use_lnL=True
        log_int_val, log_var, eff_samp, dict_return =  self.integrate_log(func, **kwargs)  # pass it on, easier than mixed coding
        if use_lnL:
          self._rvs['integrand'] = self._rvs["log_integrand"]

        return log_int_val, log_var, eff_samp, dict_return



from __future__ import print_function

import regions
import scipy.special
import numpy as np
from matplotlib import pylab as plt
from mpl_toolkits.mplot3d import Axes3D
import scipy.interpolate as interpolate

slowTests=True
verySlowTests=True




### GROUP 0: Circular region.  For illustration
print(" ------- Test group 0: Circular region ")

el = regions.RegionEllipse([['x',-2,1],['y',-1,1]])

print(" -- 0.1: Sample region ")
print("  Plot 1:  a circle, blue point inside, red outside ")
print("  Plot 2:  as above, in polar coordinates (in principal axes of circle) ")
# Test 1: Identify region boundary (default circle)
#   To draw an ellipse, use http://matplotlib.org/api/artist_api.html#matplotlib.patches.Ellipse
#   Related infrastructure test: express 'good' points in associated polar coordinates
if slowTests:
#    pts = el.draw(500)     # do not use the native draw function, so I can find points *outside* the region
    pts = np.random.randn(1000,2)
    pts_test = [el.inside_q(x) for x in pts]
    ptsOk = np.array([pts[k] for k, value in enumerate(pts_test) if value])
    ptsBad = np.array([pts[k] for k, value in enumerate(pts_test) if not value])
    print(len(ptsOk), pts.shape)
    fig = plt.gcf()
    plt.scatter(ptsOk[:,0], ptsOk[:,1])
    if len(ptsBad)>0:
        plt.scatter(ptsBad[:,0], ptsBad[:,1],color='r')
    fig.gca().add_artist(plt.Circle((0,0),1,fill=False))
    plt.xlabel('x')
    plt.ylabel('y')
    fig = plt.figure(2)
    ptsOk_internal = np.array(map(el.convert_global_to_internal, ptsOk))
    plt.scatter(ptsOk_internal[:,0], ptsOk_internal[:,1])
    if len(ptsBad)>0:
        ptsBad_internal = np.array(map(el.convert_global_to_internal, ptsBad))
        plt.scatter(ptsBad_internal[:,0], ptsBad_internal[:,1],color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')
    plt.show()


# Test 2: Integrate functions on this region (1; gaussian; ...), using BRUTE FORCE (cartesian+rejection) method
#    Note the integral will be independent of the region boundaries
print(" -- 0.2: Integrate over region via mcsampler: 1,2, and gaussian ")
mypi,err = el.integrate(lambda x: 1, lambda x:1.,verbose=False)
print("The following quantity should be pi, plus a small error: ",  mypi,err)  # circle of radius 1
mypi,err = el.integrate(lambda x: 1, lambda x:2.,verbose=False)
print("The following quantity should be 2pi, plus a small error: ",  mypi,err)  # circle of radius 1

if slowTests:
    sigma = 0.1
    myunit,err = el.integrate(lambda x: np.exp(-np.dot(x,x)/(2.*sigma**2))/(2*np.pi*sigma**2), lambda x:1.,verbose=False,adaptive=True)
    print("The following quantity should be 1, plus a small error: ",  myunit,err)  # gaussian centered on 0


# Test 3: Marginalize: construct posterior distributions for 1d quantitiees
print(" -- 0.3(a): Marginalized distribution [1 over circle] ")
print("  Plot 1:  cumulative area of unit circle to left of 'x, compared with exact formula' ")
weighted_samples = el.sample_and_marginalize(lambda x: x[0], lambda x: 1, lambda x:1.,verbose=False,sort_by_parameter=True)
# Construct 1d cumulative versus *parameter*.
# Should agree with 'x' for x in [-1,1] (basically, area of circle]
# Limited by finite neff
print("neff for this distribution is", np.sum(weighted_samples[:,1])/np.max(weighted_samples[:,1]))  # neff
xtmp =weighted_samples[:,0]
plt.plot(xtmp,weighted_samples[:,2],label="from samples")                                                     # Cumulative distribution in 1d (scatter)
plt.plot(xtmp,  0.5*((2*(xtmp*np.sqrt(1.-xtmp*xtmp) + np.arcsin(xtmp)))/np.pi+1),label="exact")  # Should recover the area of a circle
plt.legend()
plt.xlabel('x')
plt.ylabel('P(<x)')
plt.show()


if slowTests:
    print(" -- 0.3(b): Marginalized distribution [narrow gaussian in circle] ")
    print("  Plot 1:  cumulative area of gaussian to left of 'x, compared with exact formula' ")
    sigma = 0.1
    weighted_samples = el.sample_and_marginalize(lambda x: x[0], lambda x: np.exp(-np.dot(x,x)/(2.*sigma**2))/(2*np.pi*sigma**2), lambda x:1.,verbose=False, sort_by_parameter=True)
#    print "neff for this distribution is", np.sum(weighted_samples[:,1])/np.max(weighted_samples[:,1])  # neff
    xtmp =weighted_samples[:,0]
    plt.plot(xtmp,weighted_samples[:,2],label="from samples")                  # Cumulative distribution in 1d (scatter).  Note this should agree with the error function when sigma is small
    plt.plot(xtmp,0.5+scipy.special.erf(xtmp/np.sqrt(2)/sigma)*0.5,label="almost exact")  # Error function
    plt.legend()
    plt.xlabel('x')
    plt.ylabel('P(<x)')
    plt.show()


### GROUP 1: Elliptical regions
print(" ------- Test group 1: Elliptical  ")
el = regions.RegionEllipse([['x',-5,5],['y',-5,5]], mtx = [[2,0],[0,0.5]])

# Test 1.1: Identify region boundary (default circle)
#   To draw an ellipse, use http://matplotlib.org/api/artist_api.html#matplotlib.patches.Ellipse
if slowTests:
    pts = np.random.randn(1000,2)
    pts_test = [el.inside_q(x) for x in pts]
    ptsOk = np.array([pts[k] for k, value in enumerate(pts_test) if value])
    ptsBad = np.array([pts[k] for k, value in enumerate(pts_test) if not value])
    print(len(ptsOk))
    plt.scatter(ptsOk[:,0], ptsOk[:,1])
    plt.scatter(ptsBad[:,0], ptsBad[:,1],color='r')
    fig = plt.figure(2)
    ptsOk_internal = np.array(map(el.convert_global_to_internal, ptsOk))
    ptsBad_internal = np.array(map(el.convert_global_to_internal, ptsBad))
    plt.scatter(ptsOk_internal[:,0], ptsOk_internal[:,1])
    plt.scatter(ptsBad_internal[:,0], ptsBad_internal[:,1],color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')

    plt.show()


### GROUP 1c: Elliptical region, with nontrivial correlation
el = regions.RegionEllipse([['x',-5,5],['y',-5,5]], mtx = [[2,0.6],[0.6,0.5]])

# Test 1.1: Identify region boundary (default circle)
#   To draw an ellipse, use http://matplotlib.org/api/artist_api.html#matplotlib.patches.Ellipse
if slowTests:
    pts = np.random.randn(1000,2)
    pts_test = [el.inside_q(x) for x in pts]
    ptsOk = np.array([pts[k] for k, value in enumerate(pts_test) if value])
    ptsBad = np.array([pts[k] for k, value in enumerate(pts_test) if not value])
    print(len(ptsOk))
    plt.scatter(ptsOk[:,0], ptsOk[:,1])
    plt.scatter(ptsBad[:,0], ptsBad[:,1],color='r')
    fig = plt.figure(2)
    ptsOk_internal = np.array(map(el.convert_global_to_internal, ptsOk))
    ptsBad_internal = np.array(map(el.convert_global_to_internal, ptsBad))
    plt.scatter(ptsOk_internal[:,0], ptsOk_internal[:,1])
    plt.scatter(ptsBad_internal[:,0], ptsBad_internal[:,1],color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')
    
    plt.show()




### GROUP 2: Circular region with cutoff: Half-circle

el = regions.RegionEllipse([['x',0,1],['y',-1,1]])

# Test 2.1: Identify region boundary (default circle)
#   To draw an ellipse, use http://matplotlib.org/api/artist_api.html#matplotlib.patches.Ellipse
if slowTests:
    pts = np.random.randn(1000,2)
    pts_test = [el.inside_q(x) for x in pts]
    ptsOk = np.array([pts[k] for k, value in enumerate(pts_test) if value])
    ptsBad = np.array([pts[k] for k, value in enumerate(pts_test) if not value])
    print(len(ptsOk))
    fig = plt.gcf()
    plt.scatter(ptsOk[:,0], ptsOk[:,1])
    plt.scatter(ptsBad[:,0], ptsBad[:,1],color='r')
    fig.gca().add_artist(plt.Circle((0,0),1,fill=False))
    fig = plt.figure(2)
    ptsOk_internal = np.array(map(el.convert_global_to_internal, ptsOk))
    ptsBad_internal = np.array(map(el.convert_global_to_internal, ptsBad))
    plt.scatter(ptsOk_internal[:,0], ptsOk_internal[:,1])
    plt.scatter(ptsBad_internal[:,0], ptsBad_internal[:,1],color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')
    plt.show()

# Test 2.2: Integrate functions on this region (1; gaussian; ...), using BRUTE FORCE (cartesian+rejection) method
#    I use a region boundary to impose a hard cutoff
mypi,err = el.integrate(lambda x: 1, lambda x:1.,verbose=True)
print("The following quantity should be pi/2, plus a small error: ",  mypi,err)  # circle of radius 1



# Test 2.3: Marginalize: construct posterior distributions for 1d quantitiees: Circle and gaussian
print(" Test 2.3 : Construct CDF for area of circle to left of 'x' ")
weighted_samples = el.sample_and_marginalize(lambda x: x[0], lambda x: 1, lambda x:1.,verbose=True,sort_by_parameter=True)
print("neff for this distribution is", np.sum(weighted_samples[:,1])/np.max(weighted_samples[:,1]))  # neff
xtmp =weighted_samples[:,0]
plt.plot(weighted_samples[:,0],weighted_samples[:,2])  # Cumulative distribution in 1d (scatter)
plt.plot(xtmp,  2*(xtmp*np.sqrt(1.-xtmp*xtmp) + np.arcsin(xtmp))/np.pi)  # Should recover the area of a half-circle
plt.xlabel('x')
plt.ylabel('P(<x)')
plt.show()



### GROUP 3:  Load in 'ellipsoid.dat' file produced by the dag-generation code
#   WARNING: CONSISTENT FILES REQUIRED
#       ellipsoid.dat,
#       intrinsic_grid.dat
#       massgrid-coal
#       
#   Format:
#       header='2x2 effective Fisher matrix
#       3rd row: ellipsoid axes
#       4th row: total ellipse area, estimated physical area'
if True: #slowTests:
    print(" -- 3.1: Sample real 'ellipsoid.dat' from CME ")
    import lalsimutils
    dat = np.loadtxt("ellipsoid.dat")
    center = dat[0]
    mtx =np.array([dat[1], dat[2]])
    match_cntr = dat[-1,0]
    radius_ile = np.sqrt(2*(1-match_cntr))            # coordinate radius that ILE will scale to unity in 'intrinsic_grid.dat', below. Based on match = 1-r^2/2 in ellipsoid coords
    mtxSmaller = mtx/((1-match_cntr)*2.)              # Rescale so x.mtxSmaller.x =1 surface corresponds to x.mtx.x = (1-match)*2
    print('Center: ',center)
    print('Matrix: ',mtx)
    print('match threshold:', match_cntr)
    el = regions.RegionEllipse([['mc',center[0]-5./np.sqrt(mtx[0,0]),center[0]+5./np.sqrt(mtx[0,0])],['eta', 0.15,0.25]],mtx=mtxSmaller,center=center)
    el_larger = regions.RegionEllipse([['mc',center[0]-5./np.sqrt(mtx[0,0]),center[0] +5./np.sqrt(mtx[0,0])],['eta', 0.15,0.25]],mtx=mtx,center=center)

    # 3.1 : Draw points natively [using the 'larger' ellipsoid to draw in a sensible place]
    pts = el_larger.draw(500)
    if len(pts) == 0:
        import sys
        sys.exit(0)
    pts_test = [el.inside_q(x) for x in pts]
    ptsOk = np.array([pts[k] for k, value in enumerate(pts_test) if value])
    ptsBad = np.array([pts[k] for k, value in enumerate(pts_test) if not value])
    print(len(ptsOk))
    fig = plt.gcf()
    plt.scatter(ptsOk[:,0], ptsOk[:,1])
    plt.scatter(ptsBad[:,0], ptsBad[:,1],color='r')
    plt.xlabel('Mc (Msun)')
    plt.ylabel('eta')
    fig = plt.figure(2)
    ptsOk_internal = np.array(map(el.convert_global_to_internal, ptsOk))
    ptsBad_internal = np.array(map(el.convert_global_to_internal, ptsBad))
    plt.scatter(ptsOk_internal[:,0], ptsOk_internal[:,1])
    plt.scatter(ptsBad_internal[:,0], ptsBad_internal[:,1],color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')
    plt.show()

    # 3.2 : Import and infer coordinates from mass sample file (intrinsic_grid.dat)
    #    indx  mc eta  r th
    #    Use convert_global_to_intrinsic to reproduce it (=confirm codes consistent)
    #    WARNING: 'intrinsic_grid.dat' uses grid radii that are UNITY at the edge, which is an ARBITRARY match contour
    #    Used to prove CME and this code have the same underlying grid transformations
    print(" -- 3.2: Import intrinsic_grid.dat ")
    print("  Plot 1:  mc, eta coordinates ")
    print("  Plot 2:  intrinsic polar coordinates of ellipsoid ")
    dat_grid = np.loadtxt("intrinsic_grid.dat")
    plt.scatter(dat_grid[:,1],dat_grid[:,2], label='CME grid')
    plt.xlabel("Mc (Msun)")
    plt.ylabel("eta")
    plt.legend()
    plt.figure(2)
    plt.scatter(dat_grid[:,3],dat_grid[:,4])  # note change of radius to 'sensible' radial coordinate
    pts_intrinsic = np.array([el.convert_global_to_internal([dat_grid[k,1],dat_grid[k,2]]) for k in np.arange(len(dat_grid))])
    plt.scatter(pts_intrinsic[:,0],np.mod(pts_intrinsic[:,1], 2*np.pi),color='r')
    plt.xlabel('polar radius (ellipsoidal native)')
    plt.ylabel('polar angle (ellipsoidal native)')
    plt.show()

    # 3.3 : Import and infer coordinates from mass sample file (output of util_MassGrid.py)
    #    indx m1 m2 lnLred neff sigma_{lnLred}
    print(" -- 3.3: Import output of util_MassGrid.py ")
    print("  Test : are all points in the ellipsoid? ")
    print("  Plot 1:  mc, eta coordinates: grid provided by file (blue) and random samples (green) out to 1/2e point ")
    print("  Plot 2:  polar coordinates in inferred from grid")
    print("  Plot 3:  3d plot of same data (no interpolation)")
    print("  Plot 4:  3d plot of same data (no interpolation), polar coordinates")
    dat_grid_physical = np.loadtxt("massgrid-coal-indexed.dat")
    dat_grid = np.array(map(lambda x: [lalsimutils.mchirp(x[1],x[2]), lalsimutils.symRatio(x[1],x[2])],   dat_grid_physical))
    dat_grid_internal = np.array(map(el.convert_global_to_internal, dat_grid) )
    # Test if points in ellipsoid
    for pt in dat_grid:
        if not(el.inside_q(pt)):
            print(" Not inside ellipse! ", pt)
    # Plot
    plt.figure(1)
    plt.scatter(ptsOk[:,0], ptsOk[:,1],color='g',label="random draws from ellipse")
    plt.scatter(dat_grid[:,0],dat_grid[:,1], label='CME grid')
    plt.xlabel("$M_c$ ($M_\odot$)");    plt.ylabel("$\eta$"); plt.legend()
    plt.figure(2)
    plt.scatter(dat_grid_internal[:,0],dat_grid_internal[:,1], label='CME grid')
    plt.xlabel("r (ellipse)");    plt.ylabel("$\theta$ (ellipse)"); plt.legend()
    fig = plt.figure(3)
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(dat_grid[:,0], dat_grid[:,1], dat_grid_physical[:,3])
    fig = plt.figure(4)
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(dat_grid_internal[:,0], dat_grid_internal[:,1], dat_grid_physical[:,3])
    plt.show()

    # 3.4 : Import and infer coordinates from mass sample file; construct interpolation; do integral, with *uniform* mc,eta prior
    #        This grid need *not* be consistent with 'ellipsoid.dat' above, so we redefine the center
    # 3.4.a Interpolate off of an unstructured grid (i.e., in raw mc, eta)
    print(" -- 3.4: Import output of util_MassGrid.py, then integrate over mc eta ")
    print("  Plot 2:  Add output of interpolation")
    center = dat_grid[0]
    lnL_interp_raw = interpolate.interp2d(dat_grid[:,0], dat_grid[:,1], dat_grid_physical[:,3])
    lnL_interp =lambda x,y: lnL_interp_raw(x,y)[0] if el.inside_q(np.array([x,y])) else 0   # force scalar return value, not array. Force 0 outside ellipse
    lnL_interp_polar_raw = interpolate.interp2d(dat_grid_internal[:,0], dat_grid_internal[:,1], dat_grid_physical[:,3])
    lnL_interp_polar =lambda x,y: lnL_interp_polar_raw(x,y)[0] if (x<1 and x>=0.) else 0   # force scalar return value, not array. Force 0 outside ellipse
    print(" Sanity check: lnL at interpolated grid center ", center,  el.inside_q(center), lnL_interp(center[0], center[1]))
    mcg, etag  = np.meshgrid(np.linspace(el.llim[0], el.rlim[0], 50), np.linspace(el.llim[1], el.rlim[1],50))
    mcG = mcg.flatten()
    etaG = etag.flatten()
    plt.figure(1)
    plt.scatter(mcG,etaG,color='b',label="example of grid")
    plt.scatter(ptsOk[:,0], ptsOk[:,1],color='g',label="random draws from ellipse")
    plt.scatter(dat_grid[:,0],dat_grid[:,1], label='CME grid')
    plt.legend()
    plt.show()
    lnL_interp_vals = np.array( [ lnL_interp(mcG[k], etaG[k]) for k in np.arange(len(mcG))])
    fig = plt.figure(2)
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(mcG,etaG,lnL_interp_vals)
    ax.scatter(dat_grid[:,0], dat_grid[:,1], dat_grid_physical[:,3],color='b')
    ax.set_zlim(0,np.max(dat_grid_physical[:,3]))
    plt.show()
    lnLScale = np.float128(np.max(dat_grid_physical[:,3]))
    fnL = lambda x: np.max([1e-7,np.exp(-lnLScale+lnL_interp(x[0], x[1]))])
    val  = lnLScale + np.log(el.integrate(fnL, lambda x:1, verbose=True))    # Vectorization should be provided by low-level code
    print("Evidence value, uniform prior: Unstructured grid interpolation: ", val)

    # 3.4a.1 : Import and infer coordinates from mass sample file; construct interpolation; construct posterior, with  *uniform* mc,eta prior
    print(" -- 3.5: Import and interpolate output of util_Mass grid.py, then construct marginalized mchirp distribution")
    weighted_samples = el.sample_and_marginalize(lambda x: x[0], fnL, lambda x:1.,verbose=True,sort_by_parameter=True)
    xtmp =weighted_samples[:,0]
    plt.plot(xtmp,weighted_samples[:,2])                                                     # Cumulative distribution in 1d (scatter)
    plt.xlabel('mc')
    plt.ylabel('P(<mc)')
    plt.show()


    # 3.4.b Repeat, using structured grid interpolation


### Infrastructure tests
print(" ------- INFRASTRUCTURE: Hyper-sphere unit vectors ---------- ")
print("Hyper-sphere unit vectors. Note ordering is NOT conventional in low-d. Fix?")
print("Two dimensional unit vector test: xhat, yhat = ", regions.hyperunit([1,0]), regions.hyperunit([1,np.pi/2]))  # should be x axis, y axis
print("Three dimensional unit vector test: xhat, yhat, zhat = ", regions.hyperunit([1,np.pi/2,0]), regions.hyperunit([1,np.pi/2,np.pi/2]), regions.hyperunit([1,0,0]))
print("Four dimensional unit vector test:  = ", end=' ')
print(regions.hyperunit([1,np.pi/2,0,0]))
print(regions.hyperunit([1,np.pi/2,np.pi/2,0]))
print(regions.hyperunit([1,np.pi/2,np.pi/2,np.pi/2]))
print(regions.hyperunit([1,0,0,0]))



#! /usr/bin/env python
"""
test_like_and_samp.py:  

"""

import lalsimutils
import numpy as np
import lal
import lalsimulation as lalsim

import factored_likelihood
import factored_likelihood_test

import sys
import functools

import numpy
try:
    from matplotlib import pylab as plt
except:
    print("- no matplotlib -")

try:
    import healpy
    from lalinference.bayestar import fits as bfits
    from lalinference.bayestar import plot as bplot
except:
    print(" -no skymaps - ")


from glue.lal import Cache
from glue.ligolw import utils, lsctables, table

import ourparams
import ourio
opts,  rosDebugMessagesDictionary = ourparams.ParseStandardArguments()
factored_likelihood.rosDebugMessagesDictionary = rosDebugMessagesDictionary
lalsimutils.rosDebugMessagesDictionary            = rosDebugMessagesDictionary
print(opts)
print(rosDebugMessagesDictionary)


checkInputs = opts.plot_ShowLikelihoodVersusTime

rosUseDifferentWaveformLengths = False    
rosUseRandomTemplateStartingFrequency = False

rosUseTargetedDistance = True
rosUseStrongPriorOnParameters = False
rosShowSamplerInputDistributions = opts.plot_ShowSamplerInputs
rosShowRunningConvergencePlots = True
rosShowTerminalSampleHistograms = True
rosUseMultiprocessing= False
rosDebugCheckPriorIntegral = False
nMaxEvals = int(opts.nmax)
print(" Running at most ", nMaxEvals, " iterations")

fracThreshold = opts.points_threshold_match

theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 
tEventFiducial = 0                                                                 # relative to GPS reference


det_dict = {}
rhoExpected ={}

tWindowReference = factored_likelihood.tWindowReference
tWindowExplore =     factored_likelihood.tWindowExplore

approxSignal = lalsim.GetApproximantFromString(opts.approx)
approxTemplate = approxSignal
ampO =opts.amporder # sets which modes to include in the physical signal
Lmax = opts.Lmax # sets which modes to include
fref = opts.fref
fminWavesSignal = opts.fmin_Template  # too long can be a memory and time hog, particularly at 16 kHz
fminSNR =opts.fmin_SNR
fSample = opts.srate

if rosUseDifferentWaveformLengths: 
    fminWavesTemplate = fminWavesSignal+0.005
else:
    if rosUseRandomTemplateStartingFrequency:
         print("   --- Generating a random template starting frequency  ---- ")
         fminWavesTemplate = fminWavesSignal+5.*np.random.random_sample()
    else:
        fminWavesTemplate = fminWavesSignal

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
data_dict_time = {}
Psig = None

# Read in *coincidence* XML (overridden by injection, if present)
if opts.coinc:
    print("Loading coinc XML:", opts.coinc)
    xmldoc = utils.load_filename(opts.coinc)
    coinc_table = table.get_table(xmldoc, lsctables.CoincInspiralTable.tableName)
    assert len(coinc_table) == 1
    coinc_row = coinc_table[0]
    event_time = float(coinc_row.get_end())  # 
    event_time_gps = lal.GPSTimeNow()    # Pack as GPSTime *explicitly*, so all time operations are type-consistent
    event_time_gps.gpsSeconds = int(event_time)
    event_time_gps.gpsNanoSeconds = int(1e9*(event_time -event_time_gps.gpsSeconds))
    theEpochFiducial = event_time_gps       # really should avoid duplicate names
    print("Coinc XML loaded, event time: %s" % str(coinc_row.get_end()))
    # Populate the SNR sequence and mass sequence
    sngl_inspiral_table = table.get_table(xmldoc, lsctables.SnglInspiralTable.tableName)
    m1, m2 = None, None
    for sngl_row in sngl_inspiral_table:
        # NOTE: gstlal is exact match, but other pipelines may not be
        assert m1 is None or (sngl_row.mass1 == m1 and sngl_row.mass2 == m2)
        m1, m2 = sngl_row.mass1, sngl_row.mass2
        rhoExpected[str(sngl_row.ifo)] = sngl_row.snr  # record for comparisons later
        if rosDebugMessagesDictionary["DebugMessages"]:
            det = str(sngl_row.ifo)
            print(det, rhoExpected[det])
    m1 = m1*lal.LAL_MSUN_SI
    m2 = m2*lal.LAL_MSUN_SI
    rho2Net = 0
    for det in rhoExpected:
        rho2Net += rhoExpected[det]**2
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" Network :", np.sqrt(rho2Net))
    # Create a 'best recovered signal'
    Psig = lalsimutils.ChooseWaveformParams(
        m1=m1,m2=m2,approx=approxSignal,
        fmin = fminWavesSignal, 
        dist=factored_likelihood.distMpcRef*1e6*lal.LAL_PC_SI,    # default distance
        fref=fref, 
        tref = event_time_gps,
        ampO=ampO,
                                )  # FIXME: Parameter mapping from trigger space to search space
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" === Coinc table : estimated signal [overridden if injection] ===")
        Psig.print_params()

# Read in *injection* XML
if opts.inj:
    print("Loading injection XML:", opts.inj)
    Psig = lalsimutils.xml_to_ChooseWaveformParams_array(str(opts.inj))[opts.event_id]  # Load in the physical parameters of the injection.  
    timeWaveform = float(-lalsimutils.hoft(Psig).epoch)
    Psig.deltaF = 1./lalsimutils.nextPow2(opts.seglen)       # Frequency binning needs to account for target segment length
    theEpochFiducial = Psig.tref  # Reset
    tEventFiducial = 0               # Reset
    print(" ++ Targeting event at time ++ ", lalsimutils.stringGPSNice(Psig.tref))
    print(" +++ WARNING: ADOPTING STRONG PRIORS +++ ")
    rosUseStrongPriorOnParameters= True

# TEST THE SEGMENT LENGTH TARGET
if Psig:
    timeSegmentLength  = float(-lalsimutils.hoft(Psig).epoch)
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" Template duration : ", timeSegmentLength)
    if timeSegmentLength > opts.seglen:
        print(" +++ CATASTROPHE : You are requesting less data than your template target needs!  +++")
        print("    Requested data size: ", opts.seglen)
        print("    Template duration  : ", timeSegmentLength)
        sys.exit(0)

# TRY TO READ IN DATA: if data specified, use it and construct the detector list from it. Otherwise...
if opts.channel_name and    (opts.opt_ReadWholeFrameFilesInCache):
    for inst, chan in map(lambda c: c.split("="), opts.channel_name):
        print("Reading channel %s from cache %s" % (inst+":"+chan, opts.cache_file))
        data_dict[inst] = lalsimutils.frame_data_to_non_herm_hoff(opts.cache_file, inst+":"+chan,window="Tukey",window_beta=0.1)
        data_dict_time[inst] = lalsimutils.frame_data_to_hoft(opts.cache_file, inst+":"+chan)
        fSample = len(data_dict[inst].data.data)*data_dict[inst].deltaF
        df = data_dict[inst].deltaF
        if Psig:
            Psig.deltaF = df
        print("Frequency binning: %f, length %d" % (data_dict[inst].deltaF, len(data_dict[inst].data.data)))
        print("Sampling rate ", fSample)
if opts.channel_name and not (opts.opt_ReadWholeFrameFilesInCache):
    if Psig:
        event_time = Psig.tref
    else:
        event_time = theEpochFiducial  # For now...get from XML if that is the option
    start_pad, end_pad = opts.seglen-opts.padding, opts.padding 
    for inst, chan in map(lambda c: c.split("="), opts.channel_name):
        print("Reading channel %s from cache %s" % (inst+":"+chan, opts.cache_file))
        # FIXME: Assumes a frame file exists covering EXACTLY the needed interval!
        taper = lalsim.LAL_SIM_INSPIRAL_TAPER_STARTEND
        data_dict[inst] = lalsimutils.frame_data_to_non_herm_hoff(opts.cache_file, inst+":"+chan, start=int(event_time)-start_pad, stop=int(event_time)+end_pad,window="Tukey",window_beta=0.1)
        data_dict_time[inst] = lalsimutils.frame_data_to_hoft(opts.cache_file, inst+":"+chan, start=int(event_time)-start_pad, stop=int(event_time)+end_pad)
        fSample = len(data_dict[inst].data.data)*data_dict[inst].deltaF
        df = data_dict[inst].deltaF
        if Psig:
            Psig.deltaF =df
        print("Frequency binning: %f, length %d" % (data_dict[inst].deltaF, len(data_dict[inst].data.data)))
        print("Sampling rate ", fSample)

#        print " Sampling rate of data ", fSample

# CREATE A DEFAULT INJECTION, IF NONEMADE TO THIS POINT
analytic_signal = False
if len(data_dict) is 0:
    analytic_signal = True

    if not(Psig):
        m1 = 4*lal.LAL_MSUN_SI
        m2 = 3*lal.LAL_MSUN_SI

        Psig = lalsimutils.ChooseWaveformParams(
            m1 = m1,m2 =m2,
            fmin = fminWavesSignal, 
            fref=fref, 
            tref = theEpochFiducial + tEventFiducial,
            approx=approxSignal,
            ampO=ampO,
            radec=True, theta=1.2, phi=2.4,
            detector='H1', 
            dist=250.*1.e6*lal.LAL_PC_SI,    # move back to aLIGO distances
            deltaT=1./fSample
                                )
        timeSegmentLength  = lalsimutils.estimateWaveformDuration(Psig)
        if timeSegmentLength > opts.seglen:
            print(" +++ CATASTROPHE : You are requesting less data than your template target needs!  +++")
            print("    Requested data size: ", opts.seglen)
            print("    Template duration  : ", timeSegmentLength)
            sys.exit(0)

    if opts.signal_incl:
        Psig.incl = opts.signal_incl
    if opts.signal_psi:
        Psig.psi = opts.signal_psi
    if opts.signal_phiref:
        Psig.phi = opts.signal_phiref
    if opts.signal_tref:
        if np.abs(opts.signal_tref) < 1000:
            Psig.tref = theEpochFiducial+opts.signal_tref  
        else:
            Psig.tref = opts.signal_tref
            theEpochFiducial=opts.signal_tref

    df = lalsimutils.findDeltaF(Psig)
    Psig.print_params()
    Psig.deltaF = df
    data_dict['H1'] = lalsimutils.non_herm_hoff(Psig)
    data_dict_time['H1'] = lalsimutils.hoft(Psig)
    Psig.detector = 'L1'
    data_dict['L1'] = lalsimutils.non_herm_hoff(Psig)
    data_dict_time['L1'] = lalsimutils.hoft(Psig)
    Psig.detector = 'V1'
    data_dict['V1'] = lalsimutils.non_herm_hoff(Psig)
    data_dict_time['V1'] = lalsimutils.hoft(Psig)

# Reset origin of time, if required
if opts.force_gps_time:
    print(" +++ USER HAS OVERRIDDEN FIDUCIAL EPOCH +++ ")
    print("  The zero of time (and the region to be windowed) will be changed; you had better know what you are doing.  ")
    print("    original " ,theEpochFiducial)
    print("    new      ", opts.force_gps_time)
    theEpochFiducial = lal.GPSTimeNow()
    theEpochFiducial.gpsSeconds = int(opts.force_gps_time)
    theEpochFiducial.gpsNanoSeconds =  int(1e9*(opts.force_gps_time - int(opts.force_gps_time)))

# TODO: Read PSD from XML
psd_dict = {}
if not(opts.psd_file):
    analyticPSD_Q = True # For simplicity, using an analytic PSD
    for det in data_dict.keys():
        psd_dict[det] = lalsim.SimNoisePSDaLIGOZeroDetHighPower   #lal.LIGOIPsd    # Preserves key equality in data_dict , psd_dict

else:
    analyticPSD_Q = False # For simplicity, using an analytic PSD
    detectors = data_dict.keys()
    df = data_dict[detectors[0]].deltaF
    for det in detectors:
        print("Reading PSD for instrument %s from %s" % (det, opts.psd_file))
        psd_dict[det] = lalsimutils.pylal_psd_to_swig_psd(lalsimutils.get_psd_series_from_xmldoc(opts.psd_file, det))
        psd_dict[det] = lalsimutils.regularize_swig_psd_series_near_nyquist(psd_dict[det], 80) # zero out 80 hz window near nyquist
        psd_dict[det] =  lalsimutils.enforce_swig_psd_fmin(psd_dict[det], fminSNR)           # enforce fmin at the psd level, HARD CUTOFF
        tmp = psd_dict[det].data.data
        print("Sanity check reporting  : min is ", np.min(tmp[np.nonzero(tmp)]), " and maximum is ", np.max(psd_dict[det].data.data))
        deltaF = data_dict[det].deltaF
        # remember the PSD is one-sided, but h(f) is two-sided. The lengths are not equal.
        psd_dict[det] = lalsimutils.extend_swig_psd_series_to_sampling_requirements(psd_dict[det], df, df*(len(data_dict[det].data.data)/2))
        print("Post-extension the new PSD has 1/df = ", 1./psd_dict[det].deltaF, " (data 1/df = ", 1./deltaF, ") and length ", len(psd_dict[det].data.data))
        tmp = psd_dict[det].data.data
        print("Post-extension sanity check reporting  : min is ", np.min(tmp[np.nonzero(tmp)]), " and maximum is ", np.max(psd_dict[det].data.data))


### MAIN POINT OF TEST
#  Plot data vs time
detectors = data_dict.keys()
print(detectors, data_dict_time.keys())
if True: # checkInputs:   # Disable until I fix the timing issue
    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P=Psig.copy()
    P.tref = Psig.tref
    print("Template target time ", float(P.tref))
    print("Data start time", float(data_dict_time[det].epoch))
    print("Zero of time ", float(theEpochFiducial))
    for det in detectors:
        P.detector=det 
        hT = lalsimutils.hoft(P)
        tvals = float(hT.epoch - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))  
        plt.figure(1)
        nSkip = 1
        plt.plot(tvals[::nSkip], np.real(hT.data.data[::nSkip]),label='template:'+det)

        tvals = float(data_dict_time[det].epoch - theEpochFiducial) + data_dict_time[det].deltaT*np.arange(len(data_dict_time[det].data.data))  # Not correct timing relative to zero - FIXME
        plt.plot(tvals[::nSkip],  np.real(data_dict_time[det].data.data[::nSkip]),label='data:'+det,linestyle='--')

    plt.legend()
    plt.xlabel("t(s), based on zero at "+lalsimutils.stringGPSNice(theEpochFiducial))
    plt.ylabel('h(t)')
    plt.xlim(-0.5, 0.5)  # focus on a narrow region near the peak
    plt.show()



#! /usr/bin/env python
#
# GOAL
#   - read in parameter XML
#   - assess parameter covariance (in some named parameters)
#   - generate random numbers, based on that covariance, to add to those parameters, 'puffing up' the distribution
#
# USAGE
#   - tries to follow util_CIP_GC

import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import scipy
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import functools
import itertools


from igwn_ligolw import lsctables, utils, ligolw
lsctables.use_in(ligolw.LIGOLWContentHandler)


parser = argparse.ArgumentParser()
parser.add_argument("--inj-file", help="Name of XML file")
parser.add_argument("--inj-file-out", default="output-randomized", help="Name of XML file")
parser.add_argument("--approx-output",default="SEOBNRv2", help="approximant to use when writing output XML files.")
parser.add_argument("--fref",default=20,type=float, help="Reference frequency used for spins in the ILE output.  (Since I usually use SEOBNRv3, the best choice is 20Hz)")
parser.add_argument("--fmin",type=float,default=20)
parser.add_argument("--parameter", action='append', help="Parameters altered")
parser.add_argument("--parameter-range", action='append', help="range")
parser.add_argument("--mc-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--eta-range",default=None,help="Eta range. Important if we have a BNS or other item that has a strong constraint.")
parser.add_argument("--mtot-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
opts=  parser.parse_args()

# Extract parameter names
coord_names = opts.parameter # Used  in fit
param_ranges = map(eval,opts.parameter_range)
if coord_names is None:
    sys.exit(0)
print(coord_names)


# Load data
P_list = lalsimutils.xml_to_ChooseWaveformParams_array(opts.inj_file)

# extract parameters to measure the co
dat_out = []
for P in P_list:
    line_out = np.zeros(len(coord_names))
    for x in np.arange(len(coord_names)):
        fac=1
        if coord_names[x] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        line_out[x] = P.extract_param(coord_names[x])/fac
        
    dat_out.append(line_out)

# Scale out physical mass

# relabel data
dat_out = np.array(dat_out)
X =dat_out[:,0:len(coord_names)]
for indx in np.arange(len(coord_names)):
    range_here= param_ranges[indx]
    X[:,indx] = np.random.uniform(low=range_here[0],high=range_here[1],size=len(X))

# Measure covariance matrix and generate random errors

# Sanity check parameters
for indx in np.arange(len(coord_names)):
    if coord_names[indx] == 'eta':
        X[:,indx] = np.minimum(X[:,indx], 0.25)
        X[:,indx] = np.maximum(X[:,indx], 0.01)
    if coord_names[indx] == 's1z' or coord_names[indx]=='s2z':
        X[:,indx] = np.minimum(X[:,indx], 0.99)
        X[:,indx] = np.maximum(X[:,indx], -0.99)



# Copy parameters back in.  MAKE SURE THIS IS POSSIBLE
for indx_P in np.arange(len(P_list)):
    for indx in np.arange(len(coord_names)):
        fac=1
        if coord_names[indx] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        P_list[indx_P].assign_param( coord_names[indx], X[indx_P,indx]*fac)


# Export
lalsimutils.ChooseWaveformParams_array_to_xml(P_list,fname=opts.inj_file_out,fref=P.fref)

    




"""
ourparams.py
   - Intended to hold all standard arguments, either in 
      - default command line settings OR 
      - via  hardcoded defaults
   - Provides command line parsing
   - Provides explicit settings for sampler parameters
"""

import argparse
import lalsimulation as lalsim
import RIFT.lalsimutils as lalsimutils
import lal

from igwn_ligolw import utils, lsctables, table

try:
    hasNR=True
    import NRWaveformCatalogManager3 as nrwf
except:
    hasNR=False
    print(" - no NR waveforms - ")

rosDebugMessagesDictionary = {}   # Mutable after import (passed by reference). Not clear if it can be used by caling routines
                                                  # BUT if every module has a `PopulateMessagesDictionary' module, I can set their internal copies


def ParseStandardArguments():
    global rosDebugMessagesDictionary

    rosDebugMessagesDictionary["Test"] = True
    rosDebugMessagesDictionary["DebugMessages"] = False
    rosDebugMessagesDictionary["DebugMessagesLong"] = False
    

    parser = argparse.ArgumentParser()
#    print " === Loading options === "
#    print "  ++ NOTE OPTIONS MUST BE IMPLEMENTED BY EACH ROUTINE INDIVIDUALLY - NOT YET STANDARDIZED  ++ "
    # Options.  Try to be consistent with lalinference

    # General parameters
    parser.add_argument("--Niter", dest='nmax',default=100000,type=int,help="Number of iterations")
    parser.add_argument("--Neff", dest='neff', default=100,type=int,help="Target number of effective samples")
    parser.add_argument("--Nskip", dest='nskip', default=2000,type=int,help="How often MC progress is reported (in verbose mode)")
    parser.add_argument("--convergence-tests-on", default=False,action='store_true')
    parser.add_argument("--d-max",default=10000,type=float,help="Maximum distance in Mpc")

    # Likelihood functions
    parser.add_argument("--LikelihoodType_raw",default=False,action='store_true')
    parser.add_argument("--LikelihoodType_MargPhi",default=False,action='store_true',help="Deprecated/disabled")
    parser.add_argument("--LikelihoodType_MargT",default=False,action='store_true',help="Deprecated/disabled")
    parser.add_argument("--LikelihoodType_MargTdisc",default=False,action='store_true',help="Deprecated/disabled")
    parser.add_argument("--LikelihoodType_MargTdisc_array",default=True,action='store_true',help="Default")
    parser.add_argument("--LikelihoodType_MargTdisc_array_vector",default=False,action='store_true',help="Use matrix operations to compute likelihood. Discrete....ViaArray is called. Inputs are scalars")
    parser.add_argument("--LikelihoodType_vectorized",default=False,action='store_true',help="Use matrix operations to compute likelihood. Discrete....ViaArrayVector is called. Inputs are vectors")
    parser.add_argument("--LikelihoodType_vectorized_noloops",default=False,action='store_true',help="Use matrix operations to compute likelihood. Discrete....ViaArrayVectorNoLoop is called. Inputs are vectors.  GPU version on this branch")
    parser.add_argument("--adapt-parameter", action='append',help = "Adapt in this parameter (ra, dec, tref, incl,dist,phi,psi)")
    parser.add_argument( "--adapt-beta", type=float,default=1)
    parser.add_argument("--adapt-adapt",action='store_true',help="Adapt the tempering exponent")
    parser.add_argument("--adapt-log",action='store_true',help="Use a logarithmic tempering exponent")
    parser.add_argument("--adapt-mix", type=float,default=0.1)
    parser.add_argument("--no-adapt-distance",  default=False, action='store_true')
    parser.add_argument("--no-adapt-sky", default=False,action='store_true')
    # Infrastructure choices:
    parser.add_argument("--skip-interpolation",dest="opt_SkipInterpolation",default=False,action='store_true')  # skip interpolation : saves time, but some things will crash later
    parser.add_argument("--skip-modes-less-than",dest="opt_SkipModeThreshold",default=1e-5,type=float, help="Skip modes with relative effect less than this factor times the 22 mode, in the overlap matrix 'U'. So a choice of 1e-2 roughly  modes greater 10%% of the 22 mode amplitude are included. Should be smaller than 1/snr^2.")

    # Noise model
    parser.add_argument("--psd-name", type=str, default="", help="psd name ('eval'). lal.LIGOIPsd, lalsim.SimNoisePSDiLIGOSRD, lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsimutils.Wrapper_AdvLIGOPsd, ... ")
    parser.add_argument("--psd-name-V", type=str, default="", help="psd name ('eval'). lalsim.SimNoisePSDVirgo")
    parser.add_argument("--psd-file",dest='psd_file',default=None, help="psd-file for all instruments. Assumes PSD for all instruments is provided in the file.  If None, uses an analytic PSD")
    parser.add_argument("--psd-file-singleifo",dest="psd_file_singleifo",action="append", help="psd file for one instrument. Not available for all cases")
    parser.add_argument("--psd-truncate-inverse",dest="psd_TruncateInverse",default=True,action='store_true')
    parser.add_argument("--psd-truncate-inverse-time",dest="psd_TruncateInverseTime",default=8,type=float)

#    parser.add_argument("--psd-file",dest='psd_file',default=None, help="instrument=psd-file, e.g. H1=H1_PSD.xml.gz. Can be given multiple times for different instruments. If None, uses an analytic PSD.  Implemented for compatibility with Pankow")

    # Numerical relativity file (will override approximant used as *template*)
    parser.add_argument("--NR-signal-group", default=None,help="Specific NR simulation group to use")
    parser.add_argument("--NR-signal-param", default=None,help="Parameter value")
    parser.add_argument("--NR-template-group", default=None,help="Specific NR simulation group to use")
    parser.add_argument("--NR-template-param", default=None,help="Parameter value")

    parser.add_argument("--ROM-template-group",default=None)
    parser.add_argument("--ROM-template-param",default=None)

    # Use external EOB matlab code
    parser.add_argument("--use-external-EOB",default=False,action='store_true')

    # Data or injection
    parser.add_argument("-c","--cache-file",dest='cache_file', default=None, help="LIGO cache file containing all data needed.")
    parser.add_argument("-C","--channel-name", action="append", help="instrument=channel-name, e.g. H1=FAKE-STRAIN. Can be given multiple times for different instruments.")
    parser.add_argument("--force-read-full-frames", dest='opt_ReadWholeFrameFilesInCache', action='store_true')
    parser.add_argument("--inj-xml", dest='inj', default=None,help="inspiral XML file containing injection information.")
    parser.add_argument("--event",type=int, dest="event_id", default=0,help="event ID of injection XML to use.")
    parser.add_argument("--coinc-xml", dest='coinc', help="gstlal_inspiral XML file containing coincidence information.")
    parser.add_argument("--seglen", dest='seglen', type=int, default=32, help="Minimum segment duration surrounding coing to be analyzed. (Will be rounded up to next power of 2).  ")
    parser.add_argument( "--padding", dest='padding',  type=int, default=2, help="Time window after the trigger to be included in the data segment.")
    parser.add_argument("--srate", dest='srate', type=int, default=4096, help="Sampling rate to use. Data filtered down to this rate, PSD extended up to this rate. Can conflict with stored PSD file.")

    # Sampling prior
    parser.add_argument("--sampling-prior-use-inj-sky-location",dest='opt_UseKnownSkyPosition',action='store_true')
    parser.add_argument("--sampling-prior-use-skymap",dest='opt_UseSkymap')
    parser.add_argument("--rotate-sky-coordinates", action='store_true')

    # Lower frequencies 
    parser.add_argument("--fmin-snr", dest='fmin_SNR', type=float, default=30)
    parser.add_argument("--fmin-template", dest='fmin_Template', type=float, default=35)  # slightly higher than signal, so shorter and event is offset (satisfy Evan's t_shift>0 assert)
    parser.add_argument("--fmax-snr", dest='fmax_SNR', type=float, default=2000)

    # Options to set template physics
    parser.add_argument("--fref", dest='fref', type=float, default=100.0, help="Waveform reference frequency [template]. Required, default is 100 Hz.")
    parser.add_argument("--amporder", dest='amporder', type=int, default=0, help="Amplitude order of PN waveforms")
    parser.add_argument("--Lmax", dest="Lmax", type=int, default=2, help="Lmax (>=2) to use")
    parser.add_argument( "--order",dest='order', type=int, default=-1, help="Phase order of PN waveforms")
    parser.add_argument("--spinOrder", dest='spinOrder', type=int, default=-1, help="Specify twice the PN order (e.g. 5 <==> 2.5PN) of spin effects to use, only for LALSimulation (default: -1 <==> Use all spin effects).")
    parser.add_argument("--noSpin",  action="store_false", dest="SpinQ",default=False, help="template will assume no spins.")
    parser.add_argument("--spinAligned", action="store_true", dest="SpinAlignedQ",default=False, help="template will assume spins aligned with the orbital angular momentum.")
    parser.add_argument("--approx",type=str, default="TaylorT4", help="Specify a template approximant and phase order to use.") # A tighter integration with GetStringFromApproximant and vice versa would be nice

    # Fix parameters at injected values.  Assumes injected values present
    parser.add_argument("--fix-rightascension", action='append_const', dest='fixparams',const='right_ascension',default=[])
    parser.add_argument("--fix-declination", action='append_const', dest='fixparams',const="declination")
    parser.add_argument("--fix-polarization", action='append_const', dest='fixparams',const='psi')
    parser.add_argument("--fix-distance", action='append_const', dest='fixparams',const='dist')
    parser.add_argument("--fix-time", action='append_const', dest='fixparams',const='tref')
    parser.add_argument("--fix-inclination", action='append_const', dest='fixparams',const='incl')
    parser.add_argument("--fix-phase", action='append_const', dest='fixparams',const='phi')

    parser.add_argument("--force-gps-time", dest='force_gps_time', default=None,type=numpy.float64)

    # Extra options to set verbosity level
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose", default=False)
    parser.add_argument("--super-verbose", action="store_true", dest="super_verbose", default=False)
    parser.add_argument("-q", "--quiet", action="store_false", dest="verbose", default=False)

    # Test options
    parser.add_argument("--mass1", dest="template_mass1", default=False, type=float)
    parser.add_argument("--mass2", dest="template_mass2", default=False, type=float)
    parser.add_argument("--signal-mass1", dest="signal_mass1", default=False, type=float)
    parser.add_argument("--signal-mass2", dest="signal_mass2", default=False, type=float)
    parser.add_argument("--eff-lambda", type=float, help="Value of effective tidal parameter. Optional, ignored if not given")
    parser.add_argument("--deff-lambda", type=float, help="Value of second effective tidal parameter. Optional, ignored if not given")
    parser.add_argument("--indicate-mass1", dest="indicate_mass1", default=False, type=float)
    parser.add_argument("--indicate-mass2", dest="indicate_mass2", default=False, type=float)
    parser.add_argument("--signal-fmin",dest="signal_fmin",default=30,type=float)
    parser.add_argument("--signal-fref",dest="signal_fref",default=0.,type=float, help="Override the default fref(=0) of XML injections ONLY")
    parser.add_argument("--signal-distance",dest="signal_distMpc",default=25,type=float)
    parser.add_argument("--signal-inclination", dest="signal_incl",default=False,type=float)
    parser.add_argument("--signal-polarization", dest="signal_psi",default=False,type=float)
    parser.add_argument("--signal-phase", dest="signal_phiref",default=False,type=float)
    parser.add_argument("--signal-time", dest="signal_tref",default=False,type=float,help="If a small floating point number (<1000), added to the fiducial epoch to shift the injection time. Otherwise, the GPS time of the injection; the reference time is set to that event time.")

    # Interactive plots
    parser.add_argument("--show-input-h",dest='plot_ShowH', action='store_true', default=False,help ="Show interactive plots of h(t). ")
    parser.add_argument("--show-sampler-results",dest='plot_ShowSampler', action='store_true', default=False,help ="Show interactive plots from sampler results. ")
    parser.add_argument("--show-sampler-inputs",dest='plot_ShowSamplerInputs', action='store_true', default=False,help ="Show interactive plots from sampler *inputs*. ")
    parser.add_argument("--show-likelihood-versus-time",dest="plot_ShowLikelihoodVersusTime", action='store_true',default=False,help="Show single-IFO likelihoods versus time. If injection parameters are known, also plot L(t) in the geocenter. Useful to confirm a signal has been correctly windowed at all!")
    parser.add_argument("--show-psd",dest='plot_ShowPSD',action='store_true',default=False,help="Plot PSDs, as imported and stored.")
    parser.add_argument("--no-interactive-plots",action='store_true', default=False)

    # File output and conditions controlling output
    parser.add_argument("--save-sampler-file", dest="points_file_base",default="sampler-output-file")
    parser.add_argument("--save-threshold-fraction", dest="points_threshold_match",type=float,default=0.0,help="Roughly speaking, target match for points to be returned.  In practice, all points with L> \sqrt{P}L_{max} are returned")
    parser.add_argument("--save-P",default=0.,type=float)
    parser.add_argument("--save-deltalnL",type=float, default=float("Inf"), 
                        help="Threshold on deltalnL for points preserved in output file.  Requires --output-file to be defined")
    parser.add_argument("--save-metadata", dest="force_store_metadata",action="store_true")
#    parser.add_argument("--save-metadata-file", dest="fname_metadata",default=None)
    parser.add_argument("--use-metadata", dest="force_use_metadata",action="store_true")
    parser.add_argument("--use-metadata-file", dest="fname_metadata",default=None)

    # DAG generation (only for dag-generating scripts)
    parser.add_argument("--code-dir", default=None,help="dag: Where to look for source (needed for safe submit files)")
    parser.add_argument("--n-queue", default=1,type=int, help="dag: How many jobs to queue")
    parser.add_argument("--n-intrinsic", default=100,type=int, help="dag: How many intrinsic points to sample")

    # Unused: To suck up parameters by letting me do search/replace on text files
    parser.add_argument("--unused-argument", default=None, help="Used to sop up arguments I want to replace")

    parser.add_argument("--save-no-samples",action='store_true')

    args = parser.parse_args()
    # Connect debugging options to argument parsing
    rosDebugMessagesDictionary["DebugMessages"] = args.verbose
    rosDebugMessagesDictionary["DebugMessagesLong"] = args.super_verbose
    if (rosDebugMessagesDictionary["DebugMessagesLong"]):
        rosDebugMessagesDictionary["DebugMessages"]=True

    # Error check: confirm that the desired template exists
#    print nrwf.internal_ParametersAvailable.keys()
    if hasNR and not ( args.NR_template_group in nrwf.internal_ParametersAvailable.keys()):
#        raise( nrwf.NRNoSimulation,args.NR_template_group)
        if args.NR_template_group:
            print(" ===== UNKNOWN NR PARAMETER ====== ")
            print(args.NR_template_group, args.NR_template_param)
    elif hasNR:
        if args.NR_template_param:
            args.NR_template_param = eval(args.NR_template_param) # needs to be evaluated
        if not ( args.NR_template_param in nrwf.internal_ParametersAvailable[args.NR_template_group]):
            print(" ===== UNKNOWN NR PARAMETER ====== ")
            print(args.NR_template_group, args.NR_template_param)

    return args, rosDebugMessagesDictionary

###
### Populate detector network and single-detector SNRs
###
def PopulateTriggerSNRs(opts):
    rhoExpected ={}
    # Read in *coincidence* XML (overridden by injection, if present)
    if opts.coinc:
        xmldoc = utils.load_filename(opts.coinc,contenthandler =lalsimutils.cthdler)
        coinc_table = table.get_table(xmldoc, lsctables.CoincInspiralTable.tableName)
        assert len(coinc_table) == 1
        coinc_row = coinc_table[0]
       # Populate the SNR sequence and mass sequence
        sngl_inspiral_table = table.get_table(xmldoc, lsctables.SnglInspiralTable.tableName)
        m1, m2 = None, None
        for sngl_row in sngl_inspiral_table:
            # NOTE: gstlal is exact match, but other pipelines may not be
            assert m1 is None or (sngl_row.mass1 == m1 and sngl_row.mass2 == m2)
            m1, m2 = sngl_row.mass1, sngl_row.mass2
            rhoExpected[str(sngl_row.ifo)] = sngl_row.snr  # record for comparisons later

    return rhoExpected

###
### Populate Psig from either inj xml, coinc, command-line options, or default choice:
###
def PopulatePrototypeSignal(opts):
    approxSignal = lalsim.GetApproximantFromString(opts.approx)
    approxTemplate = approxSignal
    ampO =opts.amporder # sets which modes to include in the physical signal
    Lmax = opts.Lmax # sets which modes to include
    fref = opts.fref
    fminWavesSignal = opts.fmin_Template  # too long can be a memory and time hog, particularly at 16 kHz
    fminSNR =opts.fmin_SNR
    fSample = opts.srate

    Psig = None

    # Read in *coincidence* XML (overridden by injection, if present)
    if opts.coinc:
        xmldoc = utils.load_filename(opts.coinc)
        coinc_table = table.get_table(xmldoc, lsctables.CoincInspiralTable.tableName)
        assert len(coinc_table) == 1
        coinc_row = coinc_table[0]
       # Populate the SNR sequence and mass sequence
        sngl_inspiral_table = table.get_table(xmldoc, lsctables.SnglInspiralTable.tableName)
        m1, m2 = None, None
        for sngl_row in sngl_inspiral_table:
            # NOTE: gstlal is exact match, but other pipelines may not be
            assert m1 is None or (sngl_row.mass1 == m1 and sngl_row.mass2 == m2)
            m1, m2 = sngl_row.mass1, sngl_row.mass2
        m1 = m1*lal.MSUN_SI
        m2 = m2*lal.MSUN_SI
       # Create a 'best recovered signal'
        Psig = lalsimutils.ChooseWaveformParams(
            m1=m1,m2=m2,approx=approxSignal,
            fmin = fminWavesSignal, 
            dist=factored_likelihood.distMpcRef*1e6*lal.PC_SI,    # default distance
            fref=fref, 
            ampO=ampO)  # FIXME: Parameter mapping from trigger space to search space

    # Read in *injection* XML
    if opts.inj:
        Psig = lalsimutils.xml_to_ChooseWaveformParams_array(str(opts.inj))[opts.event_id]  # Load in the physical parameters of the injection.  
        Psig.deltaT = 1./fSample   # needed if we will generate fake data from the injection xml *by this program, internally* 
        timeWaveform = lalsimutils.estimateWaveformDuration(Psig)
        Psig.deltaF = 1./lalsimutils.nextPow2(opts.seglen)       # Frequency binning needs to account for target segment length


    if  not Psig and opts.channel_name:  # If data loaded but no signal generated
            if (not opts.template_mass1) or (not opts.template_mass2) or (not opts.force_gps_time):
                print(" CANCEL: Specifying parameters via m1, m2, and time on the command line ")
            Psig = lalsimutils.ChooseWaveformParams(approx=approxSignal,
                                                    fmin = fminWavesSignal, 
                                                    dist=factored_likelihood.distMpcRef*1e6*lal.PC_SI,    # default distance
                                                    fref=fref)
            Psig.m1 = lal.MSUN_SI*opts.template_mass1
            Psig.m2 = lal.MSUN_SI*opts.template_mass2
            Psig.tref = lal.LIGOTimeGPS(0.000000000)  # Initialize as GPSTime object
            Psig.tref += opts.force_gps_time # Pass value of float into it

            
    if not(Psig):
            m1 = 4*lal.MSUN_SI
            m2 = 3*lal.MSUN_SI
            Psig = lalsimutils.ChooseWaveformParams(
                m1 = m1,m2 =m2,
                fmin = fminWavesSignal, 
                fref=fref, 
                approx=approxSignal,
                ampO=ampO
                )
    # Use forced parameters, if provided
    if opts.template_mass1:
        Psig.m1 = opts.template_mass1*lal.LAL_MSUN_SI
    if opts.template_mass2:
        Psig.m2 = opts.template_mass2*lal.LAL_MSUN_SI

    return Psig

###
### Populate standard sampler arguments
###
import RIFT.integrators.mcsampler as mcsampler
import numpy 
import functools

import RIFT.likelihood.factored_likelihood as factored_likelihood
tWindowExplore  = factored_likelihood.tWindowExplore
Tmax = numpy.max([0.05,tWindowExplore[1]]) # max ref. time
Tmin = numpy.min([-0.05,tWindowExplore[0]]) # min ref. time
Dmax = 10000 #* 1.e6 * lal.LAL_PC_SI # max distance. CAN CAUSE PROBLEMS with the CDF integrator if this limit is large (SI units). SHOULD BE OVERRIDDED
Dmin = 1. # * 1.e6 * lal.LAL_PC_SI      # min distance.  Needs to be nonzero to avoid integrator stupid problems with reconstructing the CDF

# set up bounds on parameters
# Polarization angle
psi_min, psi_max = 0, numpy.pi
# RA and dec
ra_min, ra_max = 0, 2*numpy.pi
dec_min, dec_max = -numpy.pi/2, numpy.pi/2
# Reference time
tref_min, tref_max = Tmin, Tmax
# Inclination angle
inc_min, inc_max = 0, numpy.pi
# orbital phi
phi_min, phi_max = 0, 2*numpy.pi
# distance
dist_min, dist_max = Dmin, Dmax


pFocusEachDimension = 0.03

def PopulateSamplerParameters(sampler, theEpochFiducial, tEventFiducial,distBoundGuess, Psig, opts):

    pinned_params ={}

    print(" Trying to pin parameters ", opts.fixparams)

    
    adapt_ra = (not opts.no_adapt_sky) #and ('ra' in opts.adapt_paramter)
    adapt_dec = (not opts.no_adapt_sky)# and ('dec' in opts.adapt_parameter)
    adapt_dist = (not opts.no_adapt_distance) # and ('dist' in opts.adapt_parameter)
   
    adapt_incl=False;adapt_phi=False; adapt_psi=False
    if opts.adapt_parameter:
        adapt_incl = ('incl' in opts.adapt_parameter)
        adapt_phi = 'phi' in opts.adapt_parameter
        adapt_psi = 'psi' in opts.adapt_parameter


    # Uniform sampling (in area) but nonuniform sampling in distance (*I hope*).  Auto-cdf inverse
    if not(opts.opt_UseSkymap):
        if opts.opt_UseKnownSkyPosition:
            print(" +++ USING INJECTED SKY POSITION TO NARROW THE SKY SEARCH REGION ")
            sampler.add_parameter("right_ascension", functools.partial(mcsampler.gauss_samp, Psig.phi,0.03), None, ra_min, ra_max, 
                              prior_pdf = mcsampler.uniform_samp_phase)
            sampler.add_parameter("declination", functools.partial(mcsampler.gauss_samp, Psig.theta,0.03), None, dec_min, dec_max, 
                              prior_pdf= mcsampler.uniform_samp_dec)
        else:
            sampler.add_parameter("right_ascension", mcsampler.uniform_samp_phase, None, ra_min, ra_max, 
                                      prior_pdf = mcsampler.uniform_samp_phase,
                                  adaptive_sampling= adapt_ra
                                  )
            sampler.add_parameter("declination",mcsampler.uniform_samp_dec, None, dec_min, dec_max, 
                          prior_pdf= mcsampler.uniform_samp_dec,
                                  adaptive_sampling= adapt_dec)
            if opts.fixparams.count('right_ascension') and Psig != None:
                print("  ++++ Fixing ra to injected value +++ ")
                pinned_params['right_ascension'] = Psig.phi
                #sampler.add_pinned_parameter('ra', Psig.phi )
            if opts.fixparams.count('declination')  and Psig:
                    print("  ++++ Fixing dec to injected value +++ ")
                    pinned_params['declination'] = Psig.theta
                    #sampler.add_pinned_parameter('dec', Psig.theta )
    # Override *everything* if I have a skymap used
    else:
        print("  ++++ USING SKYMAP +++ ")
        print(" ==Loading skymap==")
        print("   skymap file ", opts.opt_UseSkymap)

        # Import modules used for skymaps only here (they will be available by reference later)
        import healpy
        from lalinference.bayestar import fits as bfits
        from lalinference.bayestar import plot as bplot
    
        smap, smap_meta = bfits.read_sky_map(opts.opt_UseSkymap)
        # Modify skymap to regularize it: don't want zero probability anywhere!
        smap = 0.999 *smap + 0.001*numpy.ones(len(smap))  # 0.11% probability of being somewhere randomly on the sky
        sides = healpy.npix2nside(len(smap))
        cum_smap = numpy.cumsum(smap)

        import bisect
        def bayestar_cdf_inv(x,y):                 # it will be passed two arguments, use one
            indx = bisect.bisect(cum_smap,x) 
            th,ph = healpy.pix2ang(sides, indx)
            if rosDebugMessagesDictionary["DebugMessagesLong"]:
                print(" skymap used x->(th,ph) :", x,th,ph)
            return ph,th-numpy.pi/2.

        def bayestar_cdf_inv_vector(x,y):   # Manually vectorize, so I can insert pdb breaks
            indxVec = map(lambda z: bisect.bisect(cum_smap,z), x) 
            th, ph = healpy.pix2ang(sides,indxVec)
            return numpy.array([ph, th-numpy.pi/2.])                            # convert to RA, DEC from equatorial. Return must be pair of arrays

        def bayestar_pdf_radec(x,y):               # look up value at targted pixel 
            ra,dec = x,y
            indx = bplot._healpix_lookup(smap,  ra, dec) # note bplot takes lon (=ra, really), lat (dec). Not sure why
            return smap[indx]

        sampler.add_parameter(("right_ascension", "declination"), 
                              None, 
                              bayestar_cdf_inv_vector, (ra_min, dec_min), (ra_max, dec_max))
        sampler.prior_pdf[("right_ascension", "declination")] = numpy.vectorize(lambda x,y: 1./len(smap)) # 1/number of pixels
        sampler.pdf[("right_ascension", "declination")] =          numpy.vectorize(bayestar_pdf_radec)              # look up value at targteed pixel 
        sampler._pdf_norm[("right_ascension", "declination")] = 1


    sampler.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max,
                      prior_pdf =mcsampler.uniform_samp_psi,
                          adaptive_sampling=adapt_psi)
    if opts.fixparams.count('psi') and Psig:
        print("  ++++ Fixing psi to injected value +++ ")
        pinned_params['psi'] = Psig.psi


    def gauss_samp_withfloor(mu, std, myfloor, x):
        return 1.0/numpy.sqrt(2*numpy.pi*std**2)*numpy.exp(-(x-mu)**2/2/std**2) + myfloor
    
    sampler.add_parameter("t_ref", functools.partial(mcsampler.gauss_samp_withfloor, tEventFiducial, 0.01,0.001), None, tref_min, tref_max, 
                      prior_pdf = functools.partial(mcsampler.uniform_samp_vector, tWindowExplore[0],tWindowExplore[1]))
    if opts.fixparams.count('tref') and Psig:
        print("  ++++ Fixing time to injected value +++ ")
        pinned_params['t_ref']  = float(Psig.tref - theEpochFiducial)
        #sampler.add_pinned_parameter("tref", float(Psig.tref - theEpochFiducial) )

    # Phase (angle of L)
    sampler.add_parameter("phi_orb", functools.partial(mcsampler.uniform_samp_vector, phi_min, phi_max), None, phi_min, phi_max,
                      prior_pdf = mcsampler.uniform_samp_phase,
                          adaptive_sampling = adapt_phi)
    if opts.fixparams.count('phi') and Psig:
            print("  ++++ Fixing phi to injected value +++ ")
            pinned_params['phi_orb'] = Psig.phiref
#            sampler.add_pinned_parameter("phi", Psig.phiref )


    # Inclination (angle of L; nonspinning)
    sampler.add_parameter("inclination", functools.partial(mcsampler.cos_samp_vector), None, inc_min, inc_max,
                      prior_pdf = mcsampler.uniform_samp_theta,
                          adaptive_sampling = adapt_incl)
    if opts.fixparams.count('incl') and Psig:
        print("  ++++ Fixing incl to injected value +++ ")
        pinned_params['inclination'] = Psig.incl
#        sampler.add_pinned_parameter("incl", Psig.incl )

    # Distance
    def quadratic_samp_withfloor(rmaxQuad,rmaxFlat,pFlat,x):
            ret = 0.
            if x<rmaxQuad:
                ret+= (1-pFlat)* x*x/(3*numpy.power(rmaxQuad,3))
            if x<rmaxFlat:
                ret +=pFlat/rmaxFlat
            return  ret
    def uniform_samp_withfloor(rmaxQuad,rmaxFlat,pFlat,x):
            ret = 0.
            if x<rmaxQuad:
                ret+= (1-pFlat)/rmaxQuad
            if x<rmaxFlat:
                ret +=pFlat/rmaxFlat
            return  ret
    quadratic_samp_withfloor_vector = numpy.vectorize(quadratic_samp_withfloor, otypes=[numpy.float])
#    uniform_samp_withfloor_vector = numpy.vectorize(uniform_samp_withfloor, otypes=[numpy.float])
    uniform_samp_withfloor_vector = mcsampler.uniform_samp_withfloor_vector
    # Use an option for maximum distance
    dist_max_to_use = opts.d_max
    sampler.add_parameter("distance",
#                          functools.partial(mcsampler.quadratic_samp_vector, distBoundGuess), None, dist_min, dist_max,
#                          functools.partial(mcsampler.uniform_samp_vector,0, distBoundGuess), None, dist_min, dist_max,
                          functools.partial( uniform_samp_withfloor_vector, numpy.min([distBoundGuess,dist_max_to_use]), dist_max_to_use, 0.001), None, dist_min, dist_max_to_use,
                         prior_pdf = functools.partial(mcsampler.quadratic_samp_vector, dist_max_to_use
                                                        ),
                          adaptive_sampling = adapt_dist
                         )
#        sampler.add_parameter("dist", functools.partial(mcsampler.quadratic_samp_vector,  distBoundGuess ), None, dist_min, dist_max, prior_pdf = numpy.vectorize(lambda x: x**2/(3.*numpy.power(dist_max,3))))
    if opts.fixparams.count('dist') and Psig:
            print("  ++++ Fixing distance to injected value +++ ")
            pinned_params['distance'] = Psig.dist/(1e6*lal.LAL_PC_SI)
            #sampler.add_pinned_parameter("dist", Psig.dist/(1e6*lal.LAL_PC_SI) )
        
    return pinned_params




"""
test_precompute_noisydata.py:  Testing the likelihood evaluation with noisy, 3-ifo data.
  - Loads frame data that Chris generated (assumed in signal_hoft) into 3-detector data sets
  - Uses an assumed PSD
  - Constructs a template which *exactly* matches the known source
  - Plots the individual detector lnL_k(t) timeseries and lnL timeseries
    for the injected parameters (i.e., sky location).  Does it work?
"""
from pylal import Fr

from factored_likelihood import * # gets most things
from matplotlib import pylab as plt
import sys

#
# Set by user
#
checkResults = True
rosUseZeroNoiseCache  = True

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
data_dict_time = {}
psd_dict = {}
rhoExpected ={}
rhoManual ={}
rhoExpectedAlt ={}
analyticPSD_Q = True # For simplicity, using an analytic PSD

# Note: Injected signal has fmin = 25. We should choose a different value.
fminWaves = 30
fminSNR = 30
fSample = 4096*4  # will be reset by data sampling rate.

theEpochFiducial = lal.LIGOTimeGPS(1000000014.000000000)   # Use actual injection GPS time (assumed from trigger)
#theEpochFiducial = lal.LIGOTimeGPS(1000000000.000000000)     # Use epoch of the data
tEventFiducial = 0   #  time relative to fiducial epoch, used to identify window to look in.  Checked empirically.

detectors = ['H1', "L1", "V1"]
psd_dict['H1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['L1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['V1'] = lalsim.SimNoisePSDiLIGOSRD

# Load IFO FFTs.
# ASSUME data has same sampling rate!
if not rosUseZeroNoiseCache:
    fnameCache = 'test1.cache'
else:
    fnameCache = 'test1-0noise.cache'
data_dict['H1'] =frame_data_to_non_herm_hoff(fnameCache, "H1"+":FAKE-STRAIN")
data_dict['V1'] =frame_data_to_non_herm_hoff(fnameCache, "V1"+":FAKE-STRAIN")
data_dict['L1'] =frame_data_to_non_herm_hoff(fnameCache, "L1"+":FAKE-STRAIN")
#print data_dict['H1'].data.data[10]  # confirm data loaded
df = data_dict['H1'].deltaF  
fSample = len(data_dict['H1'].data.data)*data_dict['H1'].deltaF  # Note two-sided
print(" sampling rate of data = ", fSample)
print(" ===  Repeating SNR calculation (only valid for noiseless data) ===")
rho2NetManual = 0
for det in detectors:
    IP = ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det])
    rhoManual[det] = IP.norm(data_dict[det])
    rho2NetManual+= rhoManual[det]*rhoManual[det]
    print(det, rhoManual[det], " via raw data")


# TARGET INJECTED SIGNAL (for reference and calibration of results)
Psig = xml_to_ChooseWaveformParams_array("mdc.xml.gz")[0]  # Load in the physical parameters of the injection (=first element)
dfSig = Psig.deltaF = findDeltaF(Psig)  # NOT the full deltaf to avoid padding problems
Psig.print_params()
rho2Net = 0
for det in detectors:
    Psig.detector = det
    hT = lsu.non_herm_hoff(Psig)
    fSampleSig = len(hT.data.data)*hT.deltaF
    IP = ComplexIP(fLow=fminSNR, fNyq=fSampleSig/2,deltaF=dfSig,psd=psd_dict[det])
    rhoExpected[det] = rhoDet = IP.norm(hT)
    rho2Net += rhoDet*rhoDet
    print(det, rhoDet, "compare to", rhoExpected[det],  "; note it arrival time relative to fiducial of ", float(ComputeArrivalTimeAtDetector(det, Psig.phi,Psig.theta,Psig.tref) - theEpochFiducial))
print("Network : ", np.sqrt(rho2Net))





# PLOTTING INPUT DATA
print(" ====== Plotting input data (time domain) downsampled by x16 ==== ")
data_dict_time['H1'] = frame_data_to_hoft(fnameCache, "H1"+":FAKE-STRAIN")
dat = data_dict_time['H1'].data.data
print(len(dat), np.max(np.abs(dat)))
tvals = data_dict_time['H1'].deltaT*np.arange(len(dat))
dat = signal.decimate(dat,16)
tvals = signal.decimate(tvals,16)
plt.figure(5)
plt.plot(tvals,dat,label='h(t) [H1]')
plt.legend()
plt.xlim(0,64)
plt.draw()


# Plot the H1 data (some time)
# fvals = data_dict['H1'].deltaF* np.arange(len(data_dict['H1'].data.data))  # remember frequencies are padded from the center out
# plt.plot(fvals, np.abs(data_dict['H1'].data.data))
# plt.show()



m1 = 4*lal.LAL_MSUN_SI
m2 = 4*lal.LAL_MSUN_SI
ampO =0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
fref = 100


print(" ======= Template specified: precomputing all quantities ==========")
# Struct to hold template parameters
# Fiducial distance provided but will not be used
P =  ChooseWaveformParams(fmin=fminWaves, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         deltaT=1./fSample,
         tref=theEpochFiducial,
         dist=100*1.e6*lal.LAL_PC_SI,  # critical that this is fiducial (will be reset later)
         deltaF=df)
rholms_intp, crossTerms, rholms, epoch_post = PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict, psd_dict, Lmax, analyticPSD_Q)




#rho2Net = 400 # complete guess -- just needed to make a line
print(rho2Net)
if checkResults == True:
    print(" ======= End to end test : lnL at the injection parameters  ==========")
    lnL = FactoredLogLikelihood(Psig, rholms_intp, crossTerms, 2)
    print(" Result ", lnL, " compared to (expected network) ", rho2Net/2 , " and ", rho2NetManual/2)

    print(" ======= rholm test: Plot the lnLdata timeseries at the injection parameters (* STILL TIME OFFSET *)  ==========")
    tCenter = float(Psig.tref - theEpochFiducial)
    print(" Printing around ", tCenter, " using zero of time at ", stringGPSNice(theEpochFiducial))
    tvals = np.linspace(tCenter +tWindowReference[0],tCenter+ tWindowReference[1], fSample*(tWindowReference[1]-tWindowReference[0]))
    for det in detectors:
        lnLData = map( lambda x: SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, det), tvals)
        lnLDataEstimate = np.ones(len(tvals))*rhoExpected[det]*rhoExpected[det]
        plt.figure(1)
        tvalsPlot = tvals 
        plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
        plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2("+det+")$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is ", tEventRelative)
    plt.plot([tEventRelative,tEventRelative],[0,rho2Net], color='k',linestyle='--')  # actual time of event (vs 'triggering' time), relative to the 'zero of time' theEpochFiducial
    plt.title("lnLdata (interpolated) vs narrow time interval")
    plt.xlabel('t(s)')
    plt.ylabel('lnLdata')
    


    print(" ======= rholm test: Plot the lnL timeseries at the injection parameters ==========")
    tCenter = float(Psig.tref - theEpochFiducial)
    print(" Printing in the neighborhood of a reference time ", tCenter, "centered on ", stringGPSNice(theEpochFiducial))
    tvals = np.linspace(tCenter +tWindowReference[0],tCenter+ tWindowReference[1], fSample*(tWindowReference[1]-tWindowReference[0]))
    P = Psig.copy()
    lnL = np.zeros(len(tvals))
    for indx in np.arange(len(tvals)):
            P.tref =  theEpochFiducial+tvals[indx]
            lnL[indx] =  FactoredLogLikelihood(P, rholms_intp, crossTerms, 2)
    print(" log L scale ",np.max(lnL))
    plt.figure(1)
    tvalsPlot = tvals 
    lnLEstimate = np.ones(len(tvals))*rho2Net/2
    plt.plot(tvalsPlot, lnL,label='lnL(t)')
    plt.plot(tvalsPlot, lnLEstimate,label="$rho^2/2(net)$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
    plt.plot([tEventRelative,tEventRelative],[0,rho2Net], color='k',linestyle='--')  # actual time of event (vs 'triggering' time), relative to the 'zero of time' theEpochFiducial
    plt.title("lnL (interpolated) vs narrow time interval")
    plt.legend()
    plt.show()
